#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Polars of an airfoil 

    A Polar_Definition defines a polars 

        type      like T1 or T2
        re        like 400000
        ma        like 0.0 
        ncrit     like 7.0 
        autoRange
        specVar   like cl or alpha
        valRange  like -2, 12, 0.2

    At runtime an airfoil may have a Polar Set having some Polars

    A Polar consists out of n OpPoints holding the aerodynamic values like cd or cm 


    Object Model  

        Polar_Definition                            - defines a single Polar

        Airfoil
            |-- Polar_Set                           - manage polars of an airfoil
                -- Polar                            - a single polar  
                    |-- OpPoint                     - operating point holding aero values 

"""

import os
from copy                   import copy 
from typing                 import Tuple, override
from enum                   import StrEnum

import numpy as np

from ..base.common_utils      import * 
from ..base.math_util         import * 
from ..base.spline            import Spline1D, Spline2D

from .airfoil               import Airfoil, Airfoil_Bezier
from .airfoil               import Flap_Definition
from .xo2_driver            import Worker, file_in_use   

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


#-------------------------------------------------------------------------------
# enums   
#-------------------------------------------------------------------------------

class StrEnum_Extended (StrEnum):
    """ enum extension to get a list of all enum values"""
    @classmethod
    def values (cls):
        return [c.value for c in cls]


class var (StrEnum_Extended):

    @classmethod
    def list_small (cls):
        """ returns a small list of main polar variables"""
        l = list (cls) [:]
        l.remove(var.CDF)
        l.remove(var.XTR)
        return l


    @override
    @classmethod
    def values (cls):
        """ returns a list of all enum values"""

        # exclude cdf (friction drag) from list of values
        val_list = super().values()
        val_list.remove("cdf")
        val_list.remove("xtr")

        return val_list


    """ polar variables """
    ALPHA   = "alpha"               
    CL      = "cl"               
    CD      = "cd"               
    CDP     = "cdp"                                     # pressure drag
    CDF     = "cdf"                                     # friction drag
    GLIDE   = "cl/cd" 
    CM      = "cm"   
    RE_CALC = "Re"                                      # Reynolds number calculated for Type 2 polars
    SINK    = "sink"                                    # "cl^1.5/cd"              
    XTRT    = "xtrt"               
    XTRB    = "xtrb"    
    XTR     = "xtr"                                     # mean value of xtrt and xtrb (used by xo2)       


class polarType (StrEnum_Extended):
    """ xfoil polar types """
    T1      = "T1"
    T2      = "T2"


SPEC_ALLOWED = [var.ALPHA, var.CL]

RE_SCALE_ROUND_TO  = 5000                               # round when polar is scaled down 
MA_SCALE_ROUND_DEC = 2


#--- Aero constants ---------------------------------------------------------------------------

TEMP_DEFAULT = 15                       # default temperature in °C
ALT_DEFAULT  = 0                        # default altitude in m (sea level)


def air_rho (temp_C = TEMP_DEFAULT, alt_m = ALT_DEFAULT) -> float: 
    """ 
    calc air density ρ (rho) in kg/m³ from temperature and altitude

    Args:   
        temp_C: temperature in °C
        alt_m: altitude in m
    """

    p = 101325 * (1 - 2.25577e-5 * alt_m)**5.25588      # calc pressure p in Pa
    t = temp_C + 273.15                                 # calc absolute temperature t in K
    
    rho = p / (287.05 * t)                              # calc air density ρ (rho) in kg/m³

    return round (rho, 3)


def air_eta (temp_C = TEMP_DEFAULT) -> float:
    """ 
    calc dynamic viscosity η (eta) in Pa·s (k/m.s) from temperature

    Args:   
        temp_C: temperature in °C
    """

    t = temp_C + 273.15                                 # calc absolute temperature t in K

    eta = 1.458e-6 * t**1.5 / (t + 110.4)               # Sutherland's formula

    return round(eta, 10)


def air_ny (temp_C = TEMP_DEFAULT, alt_m = ALT_DEFAULT) -> float:
    """ 
    calc kinematic viscosity ν (nu) in m²/s from temperature and altitude

    Args:   
        temp_C: temperature in °C
        alt_m: altitude in m
    """

    rho = air_rho (temp_C, alt_m)
    eta = air_eta (temp_C)

    ny = eta / rho 

    return round(ny, 10)


# convenience constants - at 15°C, sea level

AIR_RHO     = air_rho (temp_C = TEMP_DEFAULT, alt_m = ALT_DEFAULT)  # density air kg/m³ at 15°C, sea level
AIR_ETA     = air_eta (temp_C = TEMP_DEFAULT)                       # η dynamic viscosity air Pa·s (k/m.s) at 15°C, sea level
AIR_NY      = AIR_ETA / AIR_RHO                                     # ν kinematic viscosity air m²/s at 15°C, η = ν * ρ



#--- Re, Re*sqrt(CL), v ----------------------------------------------------------------


def re_from_v (v : float, chord : float, round_to = 1000) -> float:
    """ 
    calc Re number from v (velocity)
    
    Args:   
        v: velocity in m/s
        chord: chord length in m
        round_to: if int, will round the Re number to this value
    """

    re = round (v * chord * AIR_RHO / AIR_ETA,0)

    if isinstance (round_to, int) and round_to:
        re = round (re / round_to, 0)
        re = re * round_to

    return re


def v_from_re (re : float, chord : float, round_dec = 1) -> float:
    """ 
    calc v (velocity) from Renumber

    Args:   
        re: Reynolds number
        chord: chord length in m
        round_dec: if int, will round the velocity to this decimal places
    """

    v = re * AIR_ETA / (chord * AIR_RHO)

    if isinstance (round_dec, int):
        v = round (v, round_dec)

    return v


def re_sqrt_from_load (load : float, chord : float, round_to = 1000) -> float:
    """ 
    calc Re*sqrt(CL) from load (kg/m²)
    
    Args:   
        load: load in kg/m² = 10 * g/dm²
        chord: chord length in m
        round_to: if int, will round the Re number to this value
    """

    re_sqrt_cl = chord * np.sqrt(AIR_RHO) / AIR_ETA * np.sqrt(2 * 9.81 * load)
    re_sqrt_cl = round (re_sqrt_cl,0)

    if isinstance (round_to, int) and round_to:
        re_sqrt_cl = round (re_sqrt_cl / round_to, 0)
        re_sqrt_cl = re_sqrt_cl * round_to

    return re_sqrt_cl




def load_from_re_sqrt (re_sqrt_cl : float, chord : float, round_dec = None) -> float:
    """ 
    calc load (kg/m²) from Re*sqrt(CL)
    
    Args:   
        re_sqrt_cl: Re*sqrt(CL)
        chord: chord length in m
        round_to: if int, will round the load to this value
    """

    load = (re_sqrt_cl * AIR_ETA / (np.sqrt(AIR_RHO) * chord))**2  / (2 * 9.81)

    if isinstance (round_dec, int):
        load = round (load / round_dec, 0)
 
    return load 


#------------------------------------------------------------------------------


class Polar_Definition:
    """ 
    Defines the properties of a Polar (independent of an airfoil) 

    Polar_Definition
    Airfoil 
        |--- Polar_Set 
                |--- Polar    <-- Polar_Definition

    """
    XTRIP_VLM = 0.05                            # default xtript/xtripb for VLM sims

    MAX_POLAR_DEFS  = 5                         # limit to check in App

    VAL_RANGE_ALPHA = [-4.0, 13.0, 0.3]         # default value range for alpha polar
    VAL_RANGE_CL    = [-0.2, 1.2, 0.05]

    def __init__(self, dataDict : dict = None):
        
        self._autoRange = fromDict (dataDict, "autoRange",True)
        self._valRange  = fromDict (dataDict, "valRange", self.VAL_RANGE_ALPHA)
        self._specVar   = None 
        self.set_specVar (fromDict (dataDict, "specVar",  var.ALPHA))       # it is a enum
        self._type      = None 
        self.set_type    (fromDict (dataDict, "type",     polarType.T1))    # it is a enum

        self._ncrit     = fromDict (dataDict, "ncrit",    7.0)
        self._xtript    = fromDict (dataDict, "xtript",   None)             # forced transition top side
        self._xtripb    = fromDict (dataDict, "xtripb",   None)             # forced transition bot side

        self._re        = fromDict (dataDict, "re",       400000)             
        self._ma        = fromDict (dataDict, "mach",     0.0)

        flap_dict       = fromDict (dataDict, "flap",     None)
        self._flap_def  = Flap_Definition (dataDict=flap_dict) if flap_dict else None

        self._active    = fromDict (dataDict, "active",   True)             # a polar definition can be in-active

        self._is_mandatory = False                                          #  polar needed e.g. for xo2



    def __repr__(self) -> str:
        """ nice print string polarType and Re """
        return f"<{type(self).__name__} {self.name}>"

    # --- save --------------------- 

    def _as_dict (self):
        """ returns a data dict with the parameters of self """

        d = {}
        toDict (d, "type",           str(self.type))                    # type is enum
        toDict (d, "re",             self.re) 
        toDict (d, "ma",             self.ma) 
        toDict (d, "ncrit",          self.ncrit) 
        toDict (d, "specVar",        str(self.specVar))                 # specVar is enum
        toDict (d, "autoRange",      self.autoRange) 
        toDict (d, "valRange",       self.valRange) 
        toDict (d, "active",         self.active) 

        if self._xtript is not None:
            toDict (d, "xtript",    self._xtript)
        if self._xtripb is not None:
            toDict (d, "xtripb",    self._xtripb)

        if self._flap_def:
            toDict (d, "flap", self._flap_def._as_dict ())
        return d


    def _get_label (self, polarType, re : float, ma : float, 
                    ncrit : float, 
                    xtript: float | None, xtripb: float | None,
                    flap_def : Flap_Definition): 
        """ return a label of these polar variables"""
        ncirt_str  = f" N{ncrit:.2f}".rstrip('0').rstrip('.') 
        ma_str     = f" M{ma:.2f}".rstrip('0').rstrip('.') if ma else ""
        xtript_str = f" Trt{xtript:.0%}" if xtript is not None else ""
        xtripb_str = f" Trb{xtripb:.0%}" if xtripb is not None else ""

        if flap_def:
            flap_str  = f" F{flap_def.flap_angle:.1f}".rstrip('0').rstrip('.') +"°" if flap_def else ""
            flap_str += f" H{flap_def.x_flap:.0%}" if flap_def.x_flap != 0.75 else ""
        else:
            flap_str = ""

        return f"{polarType} Re{int(re/1000)}k{ma_str}{ncirt_str}{flap_str}{xtript_str}{xtripb_str}"
    

    @property
    def active (self) -> bool:
        """ True - self is in use"""
        return self._active 
    
    def set_active (self, aBool : bool):
        self._active = aBool == True 


    @property 
    def is_mandatory (self) -> bool:
        """ is self needed e.g. for Xoptfoil2"""
        return self._is_mandatory
    
    def set_is_mandatory (self, aBool):
        self._is_mandatory = aBool == True


    @property
    def ncrit (self) -> float:
        """ ncrit of polar""" 
        return self._ncrit
    def set_ncrit (self, aVal : float): 
        if aVal is not None and (aVal > 0.0 and aVal < 20.0):
            self._ncrit = aVal 


    @property
    def xtript (self) -> float:
        """ forced transition top side 0..1"""
        return self._xtript if self._xtript is not None else 1.0

    def set_xtript (self, aVal : float):
        if aVal is None:
            self._xtript = None
        else:
            xtript = round (clip (aVal, 0.0, 1.0), 2)
            self._xtript = xtript if xtript < 1.0 else None


    @property
    def xtripb (self) -> float:
        """ forced transition bottom side 0..1"""
        return self._xtripb if self._xtripb is not None else 1.0

    def set_xtripb (self, aVal : float):
        if aVal is None:
            self._xtripb = None
        else:
            xtripb = round (clip (aVal, 0.0, 1.0), 2)
            self._xtripb = xtripb if xtripb < 1.0 else None

    @property
    def has_xtrip (self) -> bool:
        """ True if forced transition is set on top or bottom side"""
        return (self.xtript < 1.0) or (self.xtripb < 1.0)

    @property
    def is_VLM_polar (self) -> bool:
        """ True if self is a VLM polar definition (xtript and xtripb set to default VLM values)"""
        return (self._xtript == self.XTRIP_VLM) and (self._xtripb == self.XTRIP_VLM)    


    @property
    def specVar (self): 
        """ ALPHA or CL defining value range"""
        return self._specVar
    
    def set_specVar (self, aVar : var): 
        """ set specVar by string or polarType"""

        if not isinstance (aVar, var):
            try:
                aVar = var(aVar)
            except ValueError:
                raise ValueError(f"{aVar} is not a valid specVar")
            
        if aVar == var.ALPHA or aVar == var.CL: 
            self._specVar = aVar 
            if self._specVar == var.ALPHA:                                       # reset value range 
                self._valRange = self.VAL_RANGE_ALPHA
            else: 
                self._valRange = self.VAL_RANGE_CL

    @property
    def type (self) -> polarType: 
        """ polarType.T1 or T2"""
        return self._type
    
    def set_type (self, aType : polarType | str):
        """ set polar type by string or polarType"""

        if not isinstance (aType, polarType):
            try:
                aType = polarType(aType)
            except ValueError:
                raise ValueError(f"{aType} is not a valid polar type")
            
        if isinstance (aType, polarType): 
            self._type = aType 
            # set specification variable depending on polar type 
            if self.type == polarType.T1:
                self.set_specVar (var.ALPHA)
            else: 
                self.set_specVar (var.CL)

    @property
    def valRange (self) -> list[float]:
        """ value range of polar  [from, to, step]""" 
        return self._valRange  

    def set_valRange (self, aRange : list): 
        if len(aRange) ==3 : 
            self._valRange = aRange 


    @property
    def autoRange (self) -> bool:
        """ auto range mode of Worker""" 
        return self._autoRange 

    def set_autoRange (self, aBool : bool): 
        self._autoRange = aBool is True  


    @property
    def valRange_string (self) -> str: 
        """ value range something like '-4, 12, 0.1' """
        if not self.autoRange:
            return ", ".join(str(x).rstrip('0').rstrip('.')  for x in self._valRange) 
        else: 
            return f"auto range ({self.valRange_step:.2f})"

    @property
    def valRange_from (self) -> float: 
        return self._valRange[0]
    def set_valRange_from (self, aVal : float): 
        if aVal < self.valRange_to:
            self._valRange[0] = aVal

    @property
    def valRange_to (self) -> float: 
        return self._valRange[1]
    def set_valRange_to (self, aVal): 
        if aVal > self.valRange_from:
            self._valRange[1] = aVal

    @property
    def valRange_step (self) -> float: 
        """ step size of value range"""
        return self._valRange[2]
    
    def set_valRange_step (self, aVal : float):
        if self.specVar == var.ALPHA:
            aVal = clip (aVal, 0.1, 1.0)
        else: 
            aVal = clip (aVal, 0.01, 0.1)
        self._valRange[2] = aVal


    @property
    def re (self) -> float: 
        """ Reynolds number of polar - in case of Type 2 it is the Re*sqrt(cl)"""
        return self._re
    def set_re (self, re): 
        self._re = clip (re, 1000, 1e+8 - 1)

    @property
    def re_asK (self) -> int: 
        """ Reynolds number base 1000 - in case of Type 2 it is the Re*sqrt(cl)"""
        return int (self.re/1000) if self.re is not None else 0 
    def set_re_asK (self, aVal): 
        self.set_re (int(aVal) * 1000)


    @property
    def ma (self) -> float: 
        """ Mach number like 0.3"""
        return self._ma
    def set_ma (self, aMach):
        mach = aMach if aMach is not None else 0.0 
        self._ma = clip (round(mach,2), 0.0, 1.0)   


    @property
    def name (self): 
        """ returns polar name as a label  """
        return self._get_label (self.type, self.re, self.ma, self.ncrit, self._xtript, self._xtripb, self.flap_def)

    @property
    def name_long (self):
        """ returns polar extended name self represents """
        return f"{self.name}  {self.specVar}: {self.valRange_string}"    

    def name_with_v (self, chord : float):
        """ returns polar name with velocity for given chord """
        v = self.calc_v_for_chord(chord)
        return f"{self.name} | {v:.1f}m/s" if v is not None else self.name


    def is_equal_to (self, aDef: 'Polar_Definition', 
                     ignore_active=False, ignore_xtrip=False) -> bool:
        """ True if aPolarDef is equals self"""

        if isinstance (aDef, Polar_Definition):
            self_dict = self._as_dict()
            aDef_dict = aDef._as_dict()
            if ignore_active:
                self_dict.pop('active', None)
                aDef_dict.pop('active', None)
            if ignore_xtrip:
                self_dict.pop('xtript', None)
                self_dict.pop('xtripb', None)
                aDef_dict.pop('xtript', None)
                aDef_dict.pop('xtripb', None)
            return self_dict == aDef_dict
        else:
            return False

    def is_in (self, polar_defs : list['Polar_Definition']):
        """ True if self is already equal in list of polar definitions"""
        for polar_def in polar_defs:
            if self.is_equal_to (polar_def, ignore_active=True): return True 
        return False 


    @property
    def is_flapped (self) -> bool:
        """ True if self has a flap definition"""
        return isinstance (self._flap_def, Flap_Definition)
    
    def set_is_flapped (self, aBool : bool):
        if aBool: 
            self.set_flap_def (Flap_Definition())
        else: 
            self.set_flap_def (None)
    
    @property
    def flap_def (self) -> Flap_Definition:
        """ an optional flap definition of self"""
        return self._flap_def 
    
    def set_flap_def (self, aDef : Flap_Definition | None):
        self._flap_def = aDef


    def calc_v_for_chord (self, chord : float) -> float | None:
        """ 
        calc velocity for given chord length in mm based on Re
            - only for Type 1 polars 
            - rounded to 1 decimal place
        """

        if chord and self.type == polarType.T1:
            v = v_from_re (self.re, chord / 1000, round_dec=1) 
            return v 
        else:
            return None
        

#------------------------------------------------------------------------------

class Polar_Set:
    """ 
    Manage the polars of an airfoil   

    Polar_Definition

    Airfoil 
        |--- Polar_Set 
                |--- Polar    <-- Polar_Definition

    """

    instances : list ['Polar_Set']= []               # keep track of all instances created to reset 


    def __init__(self, myAirfoil: Airfoil, 
                 polar_def : Polar_Definition | list | None = None,
                 re_scale : float | None = None,
                 only_active : bool = False):
        """
        Main constructor for new polar set which belongs to an airfoil 

        Args:
            myAirfoil: the airfoil object it belongs to 
            polar_def: (list of) Polar_Definition to be added initially
            re_scale: will scale (down) all polars reynolds and mach number of self
            only_active: add only the 'active' polar definitions
        """

        self._airfoil = myAirfoil 
        self._polars = []                                   # list of Polars of self is holding

        re_scale = re_scale if re_scale is not None else 1.0 
        self._re_scale = clip (re_scale, 0.001, 100)

        self._add_polar_defs (polar_def, re_scale=self._re_scale, only_active=only_active)  # add initial polar def 


    def __repr__(self) -> str:
        """ nice representation of self """
        return f"<{type(self).__name__} of {self.airfoil}>"


    @property
    def airfoil (self) -> Airfoil: return self._airfoil

    @property
    def airfoil_pathFileName_abs (self) -> str:
        """ returns absolute path of airfoil"""
        abs_path = None
        if self.airfoil:
            abs_path = self.airfoil.pathFileName_abs

            # in case of Bezier we'll write only the .bez file 
            if self.airfoil.isBezierBased:
                abs_path = os.path.splitext(abs_path)[0] + Airfoil_Bezier.Extension

            # in case of hicks henne .dat is used 
            elif self.airfoil.isHicksHenneBased:
                abs_path = os.path.splitext(abs_path)[0] + Airfoil.Extension

        return abs_path


    def airfoil_ensure_being_saved (self):
        """ check and ensure that airfoil is saved to file (Worker needs it)"""

        if os.path.isfile (self.airfoil_pathFileName_abs) and not self.airfoil.isModified:
            pass 
        else: 
            if self.airfoil.isBezierBased:                      # for Bezier write only .bez - no dat
                self.airfoil.save(onlyShapeFile=True)
            else: 
                self.airfoil.save()
            logger.debug (f'Airfoil {self.airfoil_pathFileName_abs} saved for polar generation') 


    @property
    def polars (self) -> list ['Polar']: 
        return self._polars


    @property
    def polars_VLM (self) -> list ['Polar']: 
        """ VLM polars of self which typically have a forced transition"""
        return [polar for polar in self.polars if polar.is_VLM_polar]


    @property
    def polars_normal (self) -> list ['Polar']: 
        """ normal polars of self which are not for VLM (forced transition)"""
        return [polar for polar in self.polars if not polar.is_VLM_polar]


    @property
    def has_polars_not_loaded (self) -> bool: 
        """ are there polars which are still not lazyloadeds when async polar generation """
        
        polars_not_loaded = [polar for polar in self.polars if not polar.isLoaded] 
        return len(polars_not_loaded) > 0
        
        
    @property
    def re_scale (self) -> float:
        """ scale factor for re of polars """
        return self._re_scale


    def is_equal_to (self, polar_set: 'Polar_Set'):
        """ True if polar_set has the same polars (defs) """

        if polar_set is None:
            return False
        
        if self.re_scale != polar_set.re_scale: 
            return False 
        
        if len(self.polars) == len(polar_set.polars):
            for i, polar in enumerate (self.polars):
                if not polar.is_equal_to (polar_set.polars[i], ignore_active=False):
                    return False
        else:
            return False 
        return True 


    def ensure_polars_VLM (self):
        """ ensure that every 'normal' polar has a sister VLM polar in self """

        polars_normal = list(filter(lambda polar: not polar.is_VLM_polar, self.polars))
        polars_VLM    = self.polars_VLM

        for polar in polars_normal:
            # is there already a VLM polar for this polar def ?
            has_vlm = False
            for vlm_polar in polars_VLM:
                if polar.is_equal_to (vlm_polar, ignore_active=True, ignore_xtrip=True):
                    has_vlm = True
                    break
            if not has_vlm:
                # create VLM polar def 
                vlm_polar_def = Polar_Definition(polar._as_dict())
                vlm_polar_def.set_xtript (Polar_Definition.XTRIP_VLM)
                vlm_polar_def.set_xtripb (Polar_Definition.XTRIP_VLM)

                # add VLM polar 
                self._add_polar_defs (vlm_polar_def, only_active=False)


    #---------------------------------------------------------------

    def _add_polar_defs (self, polar_defs, 
                        re_scale :float | None = None,
                        only_active : bool = False):
        """ 
        Adds polars based on a active polar_def to self.
        The polars won't be loaded (or generated) 

        polar_defs can be a list or a single Polar_Definition

        re_scale will scale (down) reynolds and mach number of all polars 
        only_active will add only the 'active' polar definitions
        """

        if isinstance(polar_defs, list):
            polar_def_list = polar_defs
        else: 
            polar_def_list = [polar_defs]

        # create polar for each polar definition 
        polar_def : Polar_Definition
        for polar_def in polar_def_list:

            # append new polar if it is active 
            if not only_active or (only_active and polar_def.active) or polar_def.is_mandatory:

                new_polar = Polar(self, polar_def, re_scale=re_scale)

                # is there already a similar polar - remove old one 
                for polar in self.polars[:]: 
                    if polar.name == new_polar.name: 
                        polar.polar_set_detach ()
                        self.polars.remove(polar)

                self.polars.append (new_polar)


    def remove_polars (self):
        """ Removes all polars of self  """
        polar: Polar
        for polar in self.polars[:]: 
            polar.polar_set_detach ()
            self.polars.remove(polar)


    def remove_polars_VLM (self):
        """ remove all VLM polars from self """
        polar: Polar
        for polar in self.polars[:]: 
            if polar.is_VLM_polar:
                polar.polar_set_detach ()
                self.polars.remove(polar)


    def load_or_generate_polars (self, normal=True, VLM=True):
        """ 
        Either loads or (if not already exist) generate polars of myAirfoil 
            for normal and/or VVLM polars of self.
        """
        # select polars to be loaded/generated

        polars : list['Polar'] = []
        if normal:
            polars.extend (self.polars_normal)

        if VLM:
            polars.extend (self.polars_VLM)

        # load already existing polar files 

        polars_not_loaded = []

        for polar in polars: 
            if not polar.isLoaded:
                polar.load_xfoil_polar ()
                if not polar.isLoaded: 
                    polars_not_loaded.append(polar)

        # polars missing - if not already done, create polar_task for Worker to generate polar 

        if polars_not_loaded:

            self.airfoil_ensure_being_saved ()                                  # a real airfoil file needed

            all_polars_of_tasks = Polar_Task.get_polars_of_tasks ()
    
            # build polar tasks bundled for same ncrit, type, ... 

            new_tasks : list [Polar_Task] = []

            for polar in polars_not_loaded: 

                if not polar in all_polars_of_tasks:                            # ensure polar isn't already in any task 
                    taken_over = False
                    for task in new_tasks:
                        taken_over =  task.add_polar (polar)                    # try to add to existing task (same ncrit etc) 
                        if taken_over: break
                    if not taken_over:                                          # new task needed 
                        new_tasks.append(Polar_Task(polar))   

            # run all worker tasks - class Polar_Task and WatchDog will take care 

            for task in new_tasks:
                task.run ()

        return 



#------------------------------------------------------------------------------


class Polar_Point:
    """ 
    A single point of a polar of an airfoil   

    airfoil 
        --> Polar_Set 
            --> Polar   (1..n) 
                --> Polar_Point  (1..n) 
    """
    def __init__(self):
        """
        Main constructor for new opPoint 

        """
        self.spec   = var.ALPHA                         # self based on ALPHA or CL
        self.alpha : float = None
        self.cl    : float = None
        self.cd    : float = None
        self.cdp   : float = None
        self.cm    : float = None 
        self.xtrt  : float = None                       # transition top side
        self.xtrb  : float = None                       # transition bot side

        self.bubble_top : tuple = None                  # bubble top side (x_start, x_end)
        self.bubble_bot : tuple = None                  # bubble bot side (x_start, x_end)

    @property
    def cdf (self) -> float: 
        if self.cd and self.cdp:                  
            return self.cd - self.cdp                   # friction drag = cd - cdp 
        else: 
            return 0.0 

    @property
    def glide (self) -> float: 
        if self.cd and self.cl:                  
            return round(self.cl/self.cd,3)  
        else: 
            return 0.0 

    @property
    def sink (self) -> float: 
        if self.cd > 0.0 and self.cl >= 0.0:                  
            return round(self.cl**1.5 / self.cd,3)
        else: 
            return 0.0 

    @property
    def xtr (self) -> float: 
        return (self.xtrt + self.xtrb) / 2 


    def get_value (self, op_var : var) -> float:
        """ get the value of the opPoint variable with id"""

        if op_var == var.CD:
            val = self.cd
        elif op_var == var.CDP:
            val = self.cdp
        elif op_var == var.CDF:
            val = self.cdf
        elif op_var == var.CL:
            val = self.cl
        elif op_var == var.ALPHA:
            val = self.alpha
        elif op_var == var.CM:
            val = self.cm
        elif op_var == var.XTRT:
            val = self.xtrt
        elif op_var == var.XTRB:
            val = self.xtrb
        elif op_var == var.GLIDE:
            val = self.glide
        elif op_var == var.RE_CALC:
            val = None                                              # not available here 
        elif op_var == var.SINK:
            val = self.sink
        elif op_var == var.XTR:
            val = self.xtr
        else:
            raise ValueError (f"Op point variable '{op_var}' not known")
        return val 


    def set_value (self, op_var : var, val : float) -> float:
        """ set the value of the opPoint variable with var id"""

        if op_var == var.CD:
            self.cd = val
        elif op_var == var.GLIDE:
            self.cd = round_down(self.cl/val,6) if val != 0.0 else 0.0
        elif op_var == var.SINK:
            self.cd = round_down(self.cl**1.5/val,6) if val != 0.0 else 0.0
        elif op_var == var.CDP:
            self.cdp = val
        elif op_var == var.CL:
            self.cl = val
        elif op_var == var.ALPHA:
            self.alpha = val
        elif op_var == var.CM:
            self.cm = val
        elif op_var == var.XTRT:
            self.xtrt  = val
        elif op_var == var.XTRB:
            self.xtrb = val
        else:
            raise ValueError (f"Op point variable '{op_var}' not supported")

    @property
    def is_bubble_bot_turbulent_separated (self) -> bool:
        """ 
        True if bottom side has turbulent separated bubble 

        Laminar BL separates, transition happens while still separated (xtr), 
        but the flow stays separated even though it’s now turbulent and only reattaches further downstream.

        Effect: You now have a turbulent separated bubble over a longer chordwise distance. 
        That thick, separated shear layer produces a big momentum deficit and a fatter wake.

        Drag: Strongly higher—longer separated region, larger displacement thickness, 
        much more pressure drag.
        """
        if self.bubble_bot:
            x_start, x_end = self.bubble_bot
            return x_end >= min(1.0, self.xtrb + 0.02) and self.xtrb < 1.0
        else:
            return False
        
    @property
    def is_bubble_top_turbulent_separated (self) -> bool:
        """ True if top side has turbulent separated bubble """
        if self.bubble_top:
            x_start, x_end = self.bubble_top
            return x_end >= min(1.0, self.xtrt + 0.02) and self.xtrt < 1.0
        else:
            return False

#------------------------------------------------------------------------------


class Polar (Polar_Definition):
    """ 
    A single polar of an airfoil created by Worker

    Polar_Definition

    Airfoil 
        |--- Polar_Set 
                |--- Polar    <-- Polar_Definition
    """

    def __init__(self, mypolarSet: Polar_Set, 
                       polar_def : Polar_Definition = None, 
                       re_scale = 1.0):
        """
        Main constructor for new polar which belongs to a polar set 

        Args:
            mypolarSet: the polar set object it belongs to 
            polar_def: optional the polar_definition to initialize self definitions
            re_scale: will scale (down) polar reynolds and mach number of self

        """
        super().__init__()
        self._polar_set = mypolarSet
        self._re_scale  = re_scale

        self._error_reason = None                       # if error occurred during polar generation 

        self._polar_points = []                         # the single polar points of self
        self._alpha = None
        self._cl    = None
        self._cd    = None
        self._cdp   = None
        self._cdf   = None
        self._cm    = None 
        self._cd    = None 
        self._xtrt  = None
        self._xtrb  = None
        self._glide = None
        self._sink  = None
        self._re_calc = None

        if polar_def: 
            self.set_active     (polar_def.active)
            self.set_type       (polar_def.type)
            self.set_re         (polar_def.re)     
            self.set_ma         (polar_def.ma)
            self.set_ncrit      (polar_def.ncrit)
            self.set_xtript     (polar_def.xtript)
            self.set_xtripb     (polar_def.xtripb)
            self.set_autoRange  (polar_def.autoRange)
            self.set_specVar    (polar_def.specVar)
            self.set_valRange   (polar_def.valRange)

            if re_scale is not None and re_scale != 1.0:                              # scale reynolds if requested
                re_scaled = round (self.re * re_scale / RE_SCALE_ROUND_TO, 0)
                re_scaled = re_scaled * RE_SCALE_ROUND_TO
                ma_scaled = round (self.ma * re_scale,  MA_SCALE_ROUND_DEC)
                self.set_re (re_scaled)
                self.set_ma (ma_scaled)
                self._re_scale  = 1.0                         # scale is now 1.0 again

            # sanity - no polar with flap angle == 0.0 
            if polar_def.flap_def and polar_def.flap_def.flap_angle != 0.0:
                self.set_flap_def   (copy (polar_def.flap_def))

    def __repr__(self) -> str:
        """ nice print string wie polarType and Re """
        return f"<{type(self).__name__} {self.name}>"

    #--------------------------------------------------------

    @property
    def polar_set (self) -> Polar_Set: return self._polar_set
    def polar_set_detach (self):
        """ detaches self from its polar set"""
        self._polar_set = None

    @property
    def re_scale (self) -> float:
        """ scale value for reynolds number """
        return self._re_scale

    @property
    def polar_points (self) -> list [Polar_Point]:
        """ returns the sorted list of Polar_Points of self """
        return self._polar_points
        
    @property
    def isLoaded (self) -> bool: 
        """ is polar data loaded from file (for async polar generation)"""
        return len(self._polar_points) > 0 or self.error_occurred
    
    @property 
    def error_occurred (self) -> bool:
        """ True if error occurred during polar generation"""
        return self._error_reason is not None
    
    @property
    def error_reason (self) -> str:
        """ reason of error during polar generation """
        return self._error_reason

    def set_error_reason (self, aStr: str):
        self._error_reason = aStr


    @property
    def alpha (self) -> np.ndarray:
        if not np.any(self._alpha): self._alpha = self._get_values_forVar (var.ALPHA)
        return self._alpha
    
    @property
    def cl (self) -> np.ndarray:
        if not np.any(self._cl): self._cl = self._get_values_forVar (var.CL)
        return self._cl
    
    @property
    def cd (self) -> np.ndarray:
        if not np.any(self._cd): self._cd = self._get_values_forVar (var.CD)
        return self._cd
    
    @property
    def cdp (self) -> np.ndarray:
        if not np.any(self._cdp): self._cdp = self._get_values_forVar (var.CDP)
        return self._cdp
        
    @property
    def cdf (self) -> np.ndarray:
        if not np.any(self._cdf): self._cdf  = self._get_values_forVar (var.CDF)
        return self._cdf
        
    @property
    def glide (self) -> np.ndarray:
        if not np.any(self._glide): self._glide = self._get_values_forVar (var.GLIDE)
        return self._glide
    
    @property
    def sink (self) -> np.ndarray:
        if not np.any(self._sink): self._sink = self._get_values_forVar (var.SINK)
        return self._sink
    
    @property
    def re_calc (self) -> np.ndarray:
        if not np.any(self._re_calc): self._re_calc = self._get_values_forVar (var.RE_CALC)
        return self._re_calc

    @property
    def cm (self) -> np.ndarray:
        if not np.any(self._cm): self._cm = self._get_values_forVar (var.CM)
        return self._cm
    
    @property
    def xtrt (self) -> np.ndarray:
        if not np.any(self._xtrt): self._xtrt = self._get_values_forVar (var.XTRT)
        return self._xtrt
    
    @property
    def xtrb (self) -> np.ndarray:
        if not np.any(self._xtrb): self._xtrb = self._get_values_forVar (var.XTRB)
        return self._xtrb

    @property
    def xtr (self) -> np.ndarray:
        """ returns the average transition values of self """
        return (self.xtrb + self.xtrt) / 2.0


    @property
    def xtript_end_idx (self) -> int | None:
        """ 
        Last index where forced transition on top side is active.
        Forced region: [0:idx+1], Natural region: [idx+1:]
        Returns None if not defined.
        """

        if self.xtrt is None or len(self.xtrt) == 0 or self._xtript is None:
            return None

        # sanity - first xtrt value must be equal forced transition
        if not np.isclose(self.xtrt[0], self._xtript, atol=1e-5):
            return None

        # find last polar point with xtrt == forced transition
        for i, xtrt in enumerate (self.xtrt):
            if not np.isclose(xtrt, self._xtript, atol=1e-5):
                return i-1
        return None


    @property
    def xtripb_start_idx (self) -> int | None:
        """
        First index where forced transition on bottom side starts.
        Natural region: [0:idx], Forced region: [idx:]
        Returns None if not defined.
        """
        if self.xtrb is None or len(self.xtrb) == 0 or self._xtripb is None:
            return None

        # no sanity as xfoil shows bug here - sometimes last xtrb value isn't equal forced transition

        # find first polar point with xtrb == forced transition
        for i, xtrb in enumerate (self.xtrb):
            if np.isclose(xtrb, self._xtripb, atol=1e-5):
                return i
        return None


    @property
    def has_bubble_top (self) -> bool:
        """ True if bubble top side is defined in any polar point """
        return any (p.bubble_top for p in self.polar_points)        
    
    @property
    def has_bubble_bot (self) -> bool:  
        """ True if bubble bot side is defined in any polar point """
        return any (p.bubble_bot for p in self.polar_points)


    @property
    def min_cd (self) -> Polar_Point:
        """ returns a Polar_Point at min cd - or None if not valid"""
        if np.any(self.cd):
            ip = np.argmin (self.cd)
            return self.polar_points [ip]


    @property
    def max_glide (self) -> Polar_Point:
        """ returns a Polar_Point at max glide - or None if not valid"""
        if np.any(self.glide):
            ip = np.argmax (self.glide)
            return self.polar_points [ip]

    @property
    def max_cl (self) -> Polar_Point:
        """ returns a Polar_Point at max cl - or None if not valid"""
        if np.any(self.cl):
            ip = np.argmax (self.cl)
            return self.polar_points [ip]


    @property
    def min_cl (self) -> Polar_Point:
        """ returns a Polar_Point at min cl - or None if not valid"""
        if np.any(self.cl):
            ip = np.argmin (self.cl)
            return self.polar_points [ip]


    @property
    def alpha0_lr (self) -> float:
        """ 
        alpha at cl=0 from linear regression of polar of alpha range around alpha0
            Use forced transition at 0.05 on upper and lower for good results
            ! no advantage in this case to alpha0 (linear interpolation)  !
        """
        if not np.any(self.cl) or not np.any(self.alpha): return None

        # define mask range around alpha0 for linear regression
        alpha_min = self.alpha0 - 2.0
        alpha_max = self.alpha0 + 4.0
        mask = (self.alpha >= alpha_min) & (self.alpha <= alpha_max)
        if np.sum(mask) < 2: return None                     # not enough points for linear regression

        cl_lr    = self.cl[mask]
        alpha_lr = self.alpha[mask]

        # do linear regression
        a, b = np.polyfit(alpha_lr, cl_lr, 1)
        alpha0_lr = -b / a

        return round(alpha0_lr, 2)


    @property
    def alpha0 (self) -> float:
        """ 
        alpha at cl=0 from linear interpolation of polar for cl=0
            Use forced transition at 0.05 on upper and lower for good results!
        """
        if np.any(self.cl) and np.any(self.alpha):
            return self.get_interpolated (var.CL, 0.0, var.ALPHA)
        else: 
            return None


    def ofVars (self, xyVars: Tuple[var, var]):
        """ returns x,y polar of the tuple xyVars"""

        x, y = [], []
        
        if isinstance(xyVars, tuple):
            x = self._ofVar (xyVars[0])
            y = self._ofVar (xyVars[1])

            # sink polar - cut values <= 0 
            if var.SINK in xyVars: 
                i = 0 
                if var.SINK == xyVars[0]:
                    for i, val in enumerate(x):
                        if val > 0.0: break
                else: 
                    for i, val in enumerate(y):
                        if val > 0.0: break
                x = x[i:]
                y = y[i:]
        return x,y 

    # -----------------------

    def _ofVar (self, polar_var: var):

        vals = []
        if   polar_var == var.CL:
            vals = self.cl
        elif polar_var == var.CD:
            vals = self.cd
        elif polar_var == var.CDP:
            vals = self.cdp
        elif polar_var == var.CDF:
            vals = self.cdf
        elif polar_var == var.ALPHA:
            vals = self.alpha
        elif polar_var == var.GLIDE:
            vals = self.glide
        elif polar_var == var.RE_CALC:
            vals = self.re_calc
        elif polar_var == var.SINK:
            vals = self.sink
        elif polar_var == var.CM:
            vals = self.cm
        elif polar_var == var.XTRT:
            vals = self.xtrt
        elif polar_var == var.XTRB:
            vals = self.xtrb
        elif polar_var == var.XTR:
            vals = self.xtr
        else:
            raise ValueError ("Unknown polar variable: %s" % polar_var)
        return vals
    

    def _get_values_forVar (self, op_var) -> np.ndarray:
        """ copy values of var from op points to array"""

        nPoints = len(self.polar_points)
        if nPoints == 0: return np.array([]) 

        values = np.zeros (nPoints)
        for i, op in enumerate(self.polar_points):

            if op_var == var.RE_CALC:                                   # special case for re_calc    
                values[i] = self._re_calc_for_op (op)
            else:
                values[i] = op.get_value (op_var)
        return values 


    def _re_calc_for_op (self, op : Polar_Point) -> float:
        """ returns the re_calc value for a single polar point"""

        if self.type == polarType.T2:
            if op.cl and self.re:
                return self.re / np.sqrt(abs(op.cl))
            else:
                return 0.0      
        else:
            return self.re if self.re else 0.0
        

    def get_interpolated (self, xVar : var, xVal : float, yVar : var,
                          allow_outside_range = False) -> float:
        """
        Interpolates yVar in polar (xVar, yVar) - returns None if not successful
           allow_outside_range = True will return the y value at the boundaries 
        """

        if not self.isLoaded: return None

        xVals = self._ofVar (xVar)
        yVals = self._ofVar (yVar)

        # find the index in xVals which is right before x
        i = bisection (xVals, xVal)
        
        # now interpolate the y-value  
        if i < (len(xVals) - 1) and i >= 0:
            x1 = xVals[i]
            x2 = xVals[i+1]
            y1 = yVals[i]
            y2 = yVals[i+1]
            y = interpolate (x1, x2, y1, y2, xVal)
            y = round (y,5) if yVar == var.CD else round(y,3)

        elif allow_outside_range:
            y = yVals[0] if i < 0 else yVals[-1]                    # see return values of bisection

        else: 
            y = None

        return y




    def get_interpolated_point (self, xVar : var, xVal : float, allow_outside_range = False) -> Polar_Point:
        """
        Returns an interpolated Polar_Point for xVar at xVal.
            If not successful, None is returned.
        allow_outside_range = True will return the point at the boundaries"""

        if not self.isLoaded: return None

        point = Polar_Point()
        point.set_value (xVar, xVal)                            # set xVar value in point

        # do not interpolate self 
        vars =  [var.CL, var.CD, var.CDP, var.ALPHA, var.CM, var.XTRT, var.XTRB]
        if xVar in vars: vars.remove(xVar)

        # set other polar variables interpolated
        for yVar in vars:

            yVal = self.get_interpolated (xVar, xVal, yVar, allow_outside_range=allow_outside_range)

            if yVal is None:
                return None                                     # no interpolation possible     
            
            point.set_value (yVar, yVal)

        return point



    #--------------------------------------------------------
   

    def load_xfoil_polar (self):
        """ 
        Loads self from Xfoil polar file.

        If loading could be done or error occurred, isLoaded will be True 
        """

        if self.isLoaded: return 

        try: 
            # polar file existing?  - if yes, load polar
            if self.is_flapped:
                flap_angle  = self.flap_def.flap_angle 
                x_flap      = self.flap_def.x_flap
                y_flap      = self.flap_def.y_flap
                y_flap_spec = self.flap_def.y_flap_spec
            else:
                flap_angle  = None 
                x_flap      = None
                y_flap      = None
                y_flap_spec = None

            airfoil_pathFileName = self.polar_set.airfoil_pathFileName_abs
            polar_pathFileName   = Worker.get_existingPolarFile (airfoil_pathFileName, 
                                                self.type, self.re, self.ma, 
                                                self.ncrit, self._xtript, self._xtripb,
                                                flap_angle, x_flap, y_flap, y_flap_spec)

            if polar_pathFileName and not file_in_use (polar_pathFileName): 

                self._import_from_file(polar_pathFileName)
                logger.debug (f'{self} loaded for {self.polar_set.airfoil}') 

        except (RuntimeError) as exc:  

            self.set_error_reason (str(exc))                # polar will be 'loaded' with error


    def _import_from_file (self, polarPathFileName):
        """
        Read data for self from an Xfoil polar file  
        """

        opPoints = []

        BeginOfDataSectionTag = "-------"
        airfoilNameTag = "Calculated polar for:"
        reTag = "Re ="
        ncritTag = "Ncrit ="
        parseInDataPoints = 0

        fpolar = open(polarPathFileName)

        # parse all lines
        for line in fpolar:

            # scan for airfoil-name
            if  line.find(airfoilNameTag) >= 0:
                splitline = line.split(airfoilNameTag)
                airfoilname = splitline[1].strip()
            # scan for Re-Number and ncrit
            if  line.find(reTag) >= 0:
                splitline = line.split(reTag)
                splitline = splitline[1].split(ncritTag)

                re_string    = splitline[0].strip()
                splitstring = re_string.split("e")
                factor = float(splitstring[0].strip())
                Exponent = float(splitstring[1].strip())
                re = factor * (10**Exponent)

                # sanity checks 
                if self.re != re: 
                    raise RuntimeError (f"Re Number of polar ({self.re}) and of polar file ({re}) not equal")

                ncrit = float(splitline[1].strip())
                if self.ncrit != ncrit: 
                    raise RuntimeError (f"Ncrit of polar ({self.ncrit}) and of polar file ({ncrit}) not equal")
                # ncrit within file ignored ...

            # scan for start of data-section
            if line.find(BeginOfDataSectionTag) >= 0:
                parseInDataPoints = 1
            else:
                # get all Data-points from this line
                if parseInDataPoints == 1:
                    # split up line detecting white-spaces
                    splittedLine = line.split(" ")
                    # remove white-space-elements, build up list of data-points
                    dataPoints = []
                    for element in splittedLine:
                        if element != '':
                            dataPoints.append(element)
                    op = Polar_Point ()
                    op.alpha = float(dataPoints[0])
                    op.cl    = float(dataPoints[1])
                    op.cd    = float(dataPoints[2])
                    op.cdp   = float(dataPoints[3])
                    op.cm    = float(dataPoints[4])
                    op.xtrt  = float(dataPoints[5])
                    op.xtrb  = float(dataPoints[6])

                    # optional bubble start-end on top and bot 
                    if len(dataPoints) == 11:

                        bubble_def = (float(dataPoints[7]), float(dataPoints[8]))
                        op.bubble_top = bubble_def if bubble_def[0] > 0.0 and bubble_def[1] > 0.0 else None

                        bubble_def = (float(dataPoints[9]), float(dataPoints[10]))
                        op.bubble_bot = bubble_def if bubble_def[0] > 0.0 and bubble_def[1] > 0.0 else None

                    opPoints.append(op)
        fpolar.close()

        if len(opPoints) > 0: 

            self._polar_points = opPoints

        else: 
            logger.error (f"{self} - import from {polarPathFileName} failed")
            raise RuntimeError(f"Could not read polar file" )
 


#------------------------------------------------------------------------------


class Polar_Task:
    """ 
    Single Task for Worker to generate polars based on parameters
    May generate many polars having same ncrit and type    

    Polar_Definition

    Airfoil 
        |--- Polar_Set 
                |--- Polar    <-- Polar_Definition
                |--- Polar_Worker_Task
    """

    instances : list ['Polar_Task']= []                 # keep track of all instances created to reset 

    def __init__(self, polar: Polar =None):
        
        self._autoRange  = None
        self._specVar    = None
        self._valRange   = None
        self._type       = None 
        self._re         = []             
        self._ma         = []

        self._ncrit      = None
        self._xtript     = None                         # forced transition top side
        self._xtripb     = None                         # forced transition bot side

        self._flap_def   = None
        self._x_flap     = None
        self._y_flap     = None
        self._y_flap_spec= None
        self._flap_angle = []

        self._flap_def   = None

        self._nPoints    = None                         # speed up polar generation with limited coordinate points

        self._polars : list[Polar] = []                 # my polars to generate 
        self._myWorker   = None                         # Worker instance which does the job
        self._finalized  = False                        # worker has done the job  

        self._airfoil_pathFileName_abs = None           # airfoil file 

        if polar:
            self.add_polar (polar) 

        Polar_Task._add_to_instances (self) 


    def __repr__(self) -> str:
        """ nice representation of self """
        return f"<{type(self).__name__} of {self._type} Re {self._re} Ma {self._ma} Ncrit {self._ncrit} Flap {self._flap_angle}>"

    #---------------------------------------------------------------

    @classmethod
    def _add_to_instances (cls , aTask : 'Polar_Task'):
        """ add aTask to instances"""

        cls.instances.append (aTask)


    @classmethod
    def get_instances (cls) -> list ['Polar_Task']:
        """ removes finalized instances and returns list of active instances"""

        n_running   = 0 
        n_finalized = 0 

        for task in cls.instances [:]:                              # copy as we modify list 
            if task.isRunning():
                n_running += 1
            elif task._finalized:                                   # task finalized - remove from list 
                n_finalized += 1
                cls.instances.remove (task)

        if len (cls.instances):
            logger.debug (f"-- {cls.__name__} {len (cls.instances)} instances, {n_running} running, {n_finalized} finalized")

        return cls.instances


    @classmethod
    def get_polars_of_tasks (cls) -> list ['Polar']:
        """ list of all polars which are currently in tasks"""

        polars = []

        for task in cls.get_instances():
            polars.extend (task._polars)
        return polars


    @classmethod
    def get_total_n_polars_running (cls) -> int:
        """ total number of polars being generated in all tasks"""

        nPolarsRunning = 0

        for task in cls.get_instances():
            nPolarsRunning += task.n_polars_running
        return nPolarsRunning


    @classmethod
    def terminate_task_of_polar (cls, polar : Polar) -> 'Polar_Task':
        """ if polar is in a Task, terminate Task"""

        for task in cls.get_instances():
            if polar in task._polars:
                task.terminate()


    @classmethod
    def terminate_instances_except_for (cls, airfoils):
        """ terminate all polar tasks except for 'airfoil' and Designs"""

        tasks = cls.get_instances () 

        for task in tasks: 

            airfoil = task._polars[0].polar_set.airfoil             # a bit complicated to get airfoil of task 

            if (not airfoil in airfoils) and (not airfoil.usedAsDesign): 
                task.terminate()                                    # will kill process 


    #---------------------------------------------------------------

    @property
    def n_polars (self) -> int:
        """ number of polars of self should generate"""
        return len(self._polars)

    @property
    def n_polars_running (self) -> int:
        """ number of polars of self which are still running"""
        nRunning = 0
        for polar in self._polars:
            if not polar.isLoaded:
                nRunning += 1
        return nRunning
    

    def add_polar (self, polar : Polar) -> bool:
        """
        add (another) polar which fits for self (polar type, ncrit, ... are the same)
        Returns True if polar is taken over by self
        """    

        # sanity - - polar already generated and loaded 
        if polar.isLoaded: return  

        taken_over = True 
        
        if not self._re: 
            self._autoRange  = polar.autoRange
            self._specVar    = polar.specVar
            self._valRange   = polar.valRange
            self._type       = polar.type
        
            self._re         = [polar.re]             
            self._ma         = [polar.ma]

            self._ncrit      = polar.ncrit
            self._xtript     = polar._xtript                    # use instance variables to allow None
            self._xtripb     = polar._xtripb                    # use instance variables to allow None

            self._flap_def   = polar.flap_def
            self._x_flap     = polar.flap_def.x_flap      if polar.flap_def else None
            self._y_flap     = polar.flap_def.y_flap      if polar.flap_def else None
            self._y_flap_spec= polar.flap_def.y_flap_spec if polar.flap_def else None
            self._flap_angle = [polar.flap_def.flap_angle] if polar.flap_def else []

            self._polars     = [polar]
            self._airfoil_pathFileName_abs = polar.polar_set.airfoil_pathFileName_abs

        # collect all polars with same type, ncrit, xtript, xtripb, specVar, valRange 
        # to allow Worker multi-threading 
        elif  self._type==polar.type and self._ncrit == polar.ncrit and \
              self._xtript == polar._xtript and self._xtripb == polar._xtripb and \
              self._autoRange == polar.autoRange and \
              self._specVar == polar.specVar and self._valRange == polar.valRange and \
              Flap_Definition.have_same_hinge (self._flap_def, polar.flap_def):
            
            self._re.append (polar.re)
            self._ma.append (polar.ma)
            if polar.is_flapped:
                self._flap_angle.append (polar.flap_def.flap_angle)

            self._polars.append (polar)

        else: 
            taken_over = False

        return taken_over 


    def run (self):
        """ run worker to generate self polars"""

        self._myWorker = Worker ()

        try:
            self._myWorker.generate_polar (self._airfoil_pathFileName_abs, 
                        self._type, self._re, self._ma, self._ncrit, 
                        xtript=self._xtript, xtripb=self._xtripb,
                        autoRange=self._autoRange, spec=self._specVar, 
                        valRange=self._valRange, run_async=True,
                        flap_angle=self._flap_angle, x_flap=self._x_flap, y_flap=self._y_flap, 
                        y_flap_spec=self._y_flap_spec, 
                        nPoints=self._nPoints)
            logger.debug (f"{self} started")


        except Exception as exc:

            logger.warning (f"{self} - polar generation failed - error: {exc}")
            for polar in self._polars:
                polar.set_error_reason (str(exc))
            self.finalize ()


    def terminate (self):
        """ kill an active workerpolar generation """
        if self._myWorker and self.isRunning():
            logger.warning (f"terminating {self}")
            self._myWorker.terminate()
        self.finalize ()


    def finalize (self):
        """ all polars generated - worker clean up """

        if self._myWorker:
            self._myWorker.finalize ()
            self._myWorker = None 

        self._finalized = True 
        self._polars    = []


    def isRunning (self) -> bool:
        """ is worker still running"""
        return self._myWorker.isRunning() if self._myWorker else False


    def isCompleted (self) -> bool:
        """ True if all polars of self are loaded"""
        for polar in self._polars:
            if not polar.isLoaded: return False
        return True 



    def load_polars (self) -> int:
        """ 
        Tries to load new generated of self polars of Worker
            Returns number of newly loaded polars
        """

        if self.isRunning():   return 0                           # if worker is still working return 

        # get worker returncode 
        worker_returncode = self._myWorker.finished_returncode if self._myWorker else 0

        nLoaded    = 0
        for polar in self._polars:

            if not polar.isLoaded:
                if worker_returncode:
                    # set error into polar - will be 'loaded'
                    polar.set_error_reason (self._myWorker.finished_errortext)
                else: 
                    # load - if error occurs, error_reason will be set 
                    polar.load_xfoil_polar ()

                if polar.isLoaded: 
                    nLoaded += 1           

        return nLoaded



# ------------------------------------------



class Polar_Splined (Polar_Definition):
    """ 
    A single polar of an airfoil splined on basis of control points 

    Airfoil 
        --> Polar_Set 
            --> Polar   
    """

    def __init__(self, mypolarSet: Polar_Set, polar_def : Polar_Definition = None):
        """
        Main constructor for new polar which belongs to a polar set 

        Args:
            mypolarSet: the polar set object it belongs to 
            polar_def: optional the polar_definition to initialize self deinitions
        """
        super().__init__()

        self._polar_set = mypolarSet

        self._polar_points = []                     # the single oppoints of self
        self._alpha = []
        self._cl = []
        self._cd = []
        self._cm = [] 
        self._cd = [] 
        self._xtrt = []
        self._xtrb = []
        self._glide = []
        self._sink = []

        if polar_def: 
            self.set_type       (polar_def.type)
            self.set_re         (polar_def.re)
            self.set_ma         (polar_def.ma)
            self.set_ncrit      (polar_def.ncrit)
            self.set_autoRange  (polar_def.autoRange)
            self.set_specVar    (polar_def.specVar)
            self.set_valRange   (polar_def.valRange)

        self._spline : Spline2D     = None   # 2 D cubic spline representation of self

        self._x                     = None   # spline knots - x coordinates  
        self._xVar                  = None   # xVar like CL 
        self._y                     = None   # spline knots - y coordinates  
        self._yVar                  = None   # yVar like CD 

    #--------------------------------------------------------

    @property
    def polar_set (self) -> Polar_Set: 
        return self._polar_set
    def polar_set_detach (self):
        """ detaches self from its polar set"""
        self._polar_set = None

    def set_knots (self, xVar, xValues, yVar, yValues):
        """ set spline knots """
        self._x     = xValues  
        self._xVar  = xVar  
        self._y     = yValues   
        self._yVar  = yVar  

    def set_knots_from_opPoints_def (self, xyVar:tuple, opPoints_def: list):
        """ set spline knots """

        if len(opPoints_def) < 3: return            # minimum for spline 

        specVar = opPoints_def[0].specVar

        if specVar == xyVar [0]:
            self._xVar  = xyVar [0] 
            self._yVar  = xyVar [1] 
        else: 
            self._xVar  = xyVar [1] 
            self._yVar  = xyVar [0] 
        self._x  = []  
        self._y  = []

        logger.debug (f"spline x: {self._xVar}   y: {self._yVar}")

        for op in opPoints_def:  
            x,y = op.xyValues_for_xyVars ((self._xVar, self._yVar)) 
            if (x is not None) and (y is not None): 
                self._x.append (x)
                self._y.append (y)

        self.set_re (op.re)
        self.set_type (op.re_type)
        self.set_ncrit (op.ncrit)
        self.set_ma (op.ma)


    @property 
    def spline (self) -> Spline1D:
        """ spline representation of self """

        if self._spline is None: 
            if len (self._x) > 3: 
                boundary = 'notaknot'
            else: 
                boundary = "natural"
            self._spline = Spline1D (self._x, self._y, boundary=boundary)
            logger.debug (f"{self} New {boundary} spline with {len (self._x)} knots")
        return self._spline


    @property
    def opPoints (self) -> list:
        """ returns the sorted list of opPoints of self """
        return self._polar_points
    
    
    @property
    def isLoaded (self) -> bool: 
        """ is polar data available"""
        return self._x and self._y
    

    @property
    def alpha (self) -> list:
        return self._alpha
    
    @property
    def cl (self) -> list:
        return self._cl
    
    @property
    def cd (self) -> list:
        return self._cd
    
    @property
    def glide (self) -> list:
        return self._glide
    
    @property
    def sink (self) -> list:
        return self._sink
    
    @property
    def cm (self) -> list:
        return self._cm
    
    @property
    def xtrt (self) -> list:
        return self._xtrt
    
    @property
    def xtrb (self) -> list:
        return self._xtrb
    
    def ofVars (self, xyVars: Tuple[var, var]):
        """ returns x,y polar of the tuple xyVars"""

        x, y = [], []
        
        if isinstance(xyVars, tuple):
            x = self._ofVar (xyVars[0])
            y = self._ofVar (xyVars[1])

            # sink polar - cut values <= 0 
            if var.SINK in xyVars: 
                i = 0 
                if var.SINK == xyVars[0]:
                    for i, val in enumerate(x):
                        if val > 0.0: break
                else: 
                    for i, val in enumerate(y):
                        if val > 0.0: break
                x = x[i:]
                y = y[i:]
        return x,y 


    def _get_values_forVar (self, var) -> list:
        """ copy values of var from op points to list"""

        nPoints = len(self.opPoints)
        if nPoints == 0: return [] 

        values  = [0] * nPoints
        op : Polar_Point
        for i, op in enumerate(self.opPoints):
            values[i] = op.get_value (var)
        return values 


    def get_interpolated_val (self, specVar, specVal, optVar):
        """ interpolates optvar in polar (specVar, optVar)"""

        if not self.isLoaded: return None

        specVals = self._ofVar (specVar)
        optVals  = self._ofVar (optVar)

        # find the index in self.x which is right before x
        jl = bisection (specVals, specVal)
        
        # now interpolate the y-value on lower side 
        if jl < (len(specVals) - 1):
            x1 = specVals[jl]
            x2 = specVals[jl+1]
            y1 = optVals[jl]
            y2 = optVals[jl+1]
            y = interpolate (x1, x2, y1, y2, specVal)
        else: 
            y = optVals[-1]

        if optVar == var.CD:
            y = round (y,5)
        else:
            y = round(y,2) 

        return y


    #--------------------------------------------------------

    
    def generate (self):
        """ 
        create polar from spline 
        """

        u = self._get_u_distribution (50)

        # x, y = self.spline.eval (u)
        x = u 
        y = self.spline.eval (u)

        self._set_var (self._xVar, x)
        self._set_var (self._yVar, y)
            
        return 

 

    def _get_u_distribution (self, nPoints):
        """ 
        returns u with nPoints 0..1
        """

        uStart = self._x[0] # 0.0
        uEnd   = self._x[-1] # 1.0
        u = np.linspace(uStart, uEnd , nPoints) 
        return u 


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  
Helper classes, dialogs, functions for an App
"""
import os
import requests

from packaging.version      import Version                                  # has to be installed

from datetime               import date, datetime
from platformdirs           import user_config_dir, user_data_dir

from packaging.version      import Version                                  # has to be installed
from PyQt6.QtCore           import QTimer
from PyQt6.QtWidgets        import QDialogButtonBox, QWidget

from .common_utils          import Parameters
from .panels                import Dialog
from .widgets               import *

import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)




# ----------------------------------------------------------------------------

def check_or_get_initial_file (initial_file : str) -> str:
    """ check if initial file exists - otherwise get last opened file from settings """

    if initial_file:
        if os.path.isfile (initial_file):
            return initial_file
        else:
            logger.error (f"Initial file '{initial_file}' doesn't exist - using Example")
            return None

    app_settings = Settings()                                           # load app settings

    last_opened = app_settings.get('last_opened', default=None) 
    if last_opened and os.path.isfile (last_opened):
        logger.info (f"Starting on last opened file '{last_opened}'")
        return last_opened
    else:
        if last_opened:
            logger.error (f"Last opened file '{last_opened}' doesn't exist anymore")
            app_settings.delete ('last_opened', purge=True)              # remove invalid entry

    return None


# ----------------------------------------------------------------------------


class Settings (Parameters):
    """ Singleton: Handles a named app setting file with a json structure""" 

    _pathFileName = None                     # the filePath of the settings

    def __init__ (self):

        if self._pathFileName is None:
            raise RuntimeError ("Settings are not initialized")
        
        super().__init__(self._pathFileName)


    @staticmethod
    def user_data_dir (app_name, app_author = False):
        """ returns directory for app data 
        
        Args:
            app_name: name of app self will belong to 
            app_author : typical for Windows - something like 'Microsoft' of 'jxjo'  
        """

        return user_data_dir (app_name, app_author, ensure_exists=True) 


    @classmethod
    def set_file (cls, app_name, app_author = False, name_suffix=None, file_extension= '.json'):
        """ static set of the file the settings will belong to 
        
        Args:
            app_name: name of app self will belong to 
            app_author : typical for Windows - something like 'Microsoft' of 'jxjo'  
            name_suffix: ... will be appended to appName - default '_settings'      
            file_extension: ... of the settings file - default 'json'       
        """

        fileName = app_name + name_suffix + file_extension if name_suffix else app_name + file_extension

        # get directory where self is located
        settings_dir  = user_config_dir (app_name, app_author, ensure_exists=True)

        cls._pathFileName = os.path.join(settings_dir, fileName)

        logger.debug (f"Reading settings from {cls._pathFileName}")


class Run_Checker:
    """
    - Check if App runs the first time in this version 
    """

    @staticmethod
    def is_first_run (current_version : str) -> bool:
        """ returns if this is the first run of this version """

        settings = Settings()             # app settings 

        last_run_version = settings.get ("last_run_version", None)
        
        if last_run_version != current_version:
            # save current version as last run version 
            settings.set("last_run_version", current_version)
            settings.save()
            logger.info (f"First run of version {current_version}")
            return True
        else:
            return False


class Update_Checker:
    """
    - Check if update for package is available, 
    - Inform User
    """

    def __init__ (self, parent : QWidget, app_name : str, package_name : str, current_version : str):

        self._app_name = app_name
        self._package_name = package_name
        self._current_version = current_version

        self._latest_version = None
        self._settings = Settings()             # app settings 

        if self.is_newer_version_available():
            QTimer.singleShot (1000, lambda: self.show_user_info (parent))


    @property
    def latest_version (self) -> str:
        """ latest available version - None if not accessible"""
        return self._latest_version if self._latest_version else None


    def _get_pypi_latest_version (self, package_name : str) -> str:
        """ 
        API request to PYPI to get latest version as string.
            return "" if failed or not available  
        """
        latest_version = ""

        try: 
            response = requests.get(f'https://pypi.org/pypi/{package_name}/json')

            if response.status_code == 200:             # http     
                latest_version = response.json()['info']['version']
                logger.debug (f"Package {package_name} version {latest_version} found on PyPI")
            elif response.status_code == 404:
                logger.warning (f"Package {package_name} not found on PyPI")
            else:
                logger.error (f"Error {response.status_code} on accessing PyPI for package {package_name}")

            # save check date and latest version in settings
            self._settings.set("update_last_check", str(date.today()))
            self._settings.set("update_latest_version", latest_version)
            self._settings.save()

        except: 
            pass

        return latest_version


    def is_newer_version_available (self) -> bool:
        """ returns if there is a newer version on PyPI"""

        # get date of last check from settings 
        last_check_date_str = self._settings.get ("update_last_check", None)
        if last_check_date_str:

            last_check_date = datetime.strptime(last_check_date_str, '%Y-%m-%d').date()
            if last_check_date  == date.today():
                # if already checked today - get version from settings
                self._latest_version = self._settings.get ("update_latest_version", "")
                logger.debug (f"Version check on PyPI already made for today (latest version: {self._latest_version})")
            else:
                logger.debug (f"Version check on PyPI made on {last_check_date_str}")


        if self._latest_version is None: 
            self._latest_version = self._get_pypi_latest_version (self._package_name)

        if self._latest_version and self._current_version:
            is_available = Version (self._latest_version) > Version (self._current_version)
        else: 
            is_available = False

        if is_available:
            logger.info (f"New version {self._latest_version} of {self._app_name} available on PyPI")
        elif self._latest_version:
            logger.info (f"{self._app_name} is up-to-date")

        return is_available


    def show_user_info (self, parent):
        """ show info dialog about a new version"""

        dont_ask_for_version = self._settings.get ("update_dont_ask_for_version", None)

        if dont_ask_for_version != self._latest_version:

            dialog = Update_Info_Dialog (parent, self._app_name, self._package_name, self._latest_version)
            dialog.exec()

            if dialog.dont_ask_version:
                self._settings.set("update_dont_ask_for_version", self._latest_version)
                self._settings.save()



class Update_Info_Dialog (Dialog):
    """
    Inform user about a newer version of app
    """

    name = "Update Check"             # will be title 

    _width  = 420
    _height = 230 

    def __init__(self, parent, app_name : str, package_name, latest_version : str):

        self._app_name = app_name
        self._latest_version = latest_version
        self._package_name = package_name

        self._dont_ask_version = False

        super().__init__(parent)

    @property
    def dont_ask_version (self) -> bool:
        """ user doesn't want to show this dialog again"""
        return self._dont_ask_version
    
    def set_dont_ask_version (self, aBool):
        self._dont_ask_version = aBool

    def _init_layout(self):

        l = QGridLayout()

        r,c = 0, 0 
        lab = Label (l,r,c, rowSpan=4, height=30)
        icon = Icon(Icon.INFO)
        pixmap = icon.pixmap ((QSize(30, 30)))
        lab.setPixmap (pixmap)

        r,c = 0, 1 
        text =  f"There is a new version <b>{self._latest_version }</b> of {self._app_name} available.<br><br>" +   \
                f"Depending on your installation mode,<br><br>" + \
                f" - either update package with 'pip install {self._package_name} -U'<br>" + \
                f" - or download new version from <a href='https://github.com/jxjo/{self._app_name}/releases/'>GitHub" 
        lab = Label    (l,r,c, height=110, wordWrap=True, get=text)
        lab.setOpenExternalLinks(True)

        r += 1
        SpaceR (l,r)
        r += 1
        CheckBox (l,r,c, text="Don't show this message again",
                  get=self.dont_ask_version, set=self.set_dont_ask_version)

        l.setColumnMinimumWidth (0, 50)
        l.setColumnStretch (1,3)

        return l
    
    def _button_box (self) -> QDialogButtonBox:
        """ returns the QButtonBox with the buttons of self"""
        buttons = QDialogButtonBox.StandardButton.Ok 
        buttonBox = QDialogButtonBox(buttons)
        buttonBox.accepted.connect(self.accept)
        return buttonBox 


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

The different operating modes of the app like View_Mode and Modify_Mode

The modes are controlled by the Modes_Manager 

"""

from PyQt6.QtCore            import Qt, pyqtSignal, QObject, QTimer, QMargins
from PyQt6.QtWidgets         import QHBoxLayout, QMessageBox, QStackedWidget, QDialog

from .base.panels            import Container_Panel, Toaster        
from .base.app_utils         import Settings

from .model.xo2_driver       import Worker, Xoptfoil2
from .model.case             import Case_Direct_Design, Case_As_Bezier

from .ui.util_dialogs        import Airfoil_Save_Dialog
from .ui.ae_panels           import *

from .ui.xo2_dialogs         import Xo2_Select_Dialog, Xo2_New_Dialog
from .ui.xo2_panels          import *

from .app_model              import App_Model, Mode_Id


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# ------------------------------------------------------------------------------


class Data_Panel (Container_Panel):
    """ 
    Base class for data panels used in different modes. 
    - support double click to minimize/maximize
    - ensure refresh when visible
    """
    
    def __init__ (self, *args, **kwargs):
        super().__init__(*args, **kwargs)


        self.set_doubleClick (self.mode.toggle_minimized)
        self.add_hint ("Double click to minimize/maximize")

        self.app_model.sig_new_airfoil.connect (self.refresh)

    @property
    def mode (self) -> 'Mode_Abstract':
        return self._app 

    @property
    def app_model (self) -> App_Model:
        return self.dataObject   

    @override
    def refresh (self, always: bool = False):
        """ refresh panels of self - only when visible  """

        if self.isVisible() or always:
            super().refresh()


# ------------------------------------------------------------------------------


class Mode_Abstract (QObject):
    """
    Abstract base class for different application modes.

    A Mode provides the lower data panel of the UI and has methods for
        - entering and exiting the mode.
        - App level functions specific to the mode.

    It is managed by the Modes_Manager.
    """

    mode_id  : Mode_Id = None                                   # the id of the mode

    sig_exit_requested          = pyqtSignal()                  # signal to request mode and app exit
    sig_leave_requested         = pyqtSignal()                  # signal to request mode leave
    sig_switch_mode_requested   = pyqtSignal(Mode_Id, object)   # signal to request mode switch
    sig_toggle_minimized        = pyqtSignal()                  # signal to request toggle of minimized state


    def __init__(self, app_model: App_Model):
        super().__init__()

        self._app_model = app_model

        self._panel:       Data_Panel = None
        self._panel_small: Data_Panel = None

        self._stacked_panel:  QStackedWidget  = None


    def __repr__(self):
        """ nice representation of self """
        return f"<{self.__class__.__name__}>"


    def _toast_message (self, msg, toast_style = style.HINT):
        """ show toast message """
        
        Toaster.showMessage (self.stacked_panel, msg, corner=Qt.Corner.BottomLeftCorner, margin=QMargins(10, 10, 10, 5),
                             toast_style=toast_style)

    @property
    def panel(self):
        return self._panel

    @property
    def panel_small(self):
        return self._panel_small

    @property
    def stacked_panel(self) -> QStackedWidget:
        """ Get the stacked widget containing both data panels. """

        if self._stacked_panel is None:
            self._stacked_panel = QStackedWidget()
            self._stacked_panel.addWidget (self.panel)
            self._stacked_panel.addWidget (self.panel_small)

            # set initial visibility
            self.set_current_panel(refresh=False)

        return self._stacked_panel


    def on_enter(self, initial_arg = None):
        """ 
        Actions to perform when entering the mode. 
            Override in subclasses if needed. 
            Call super().on_enter() to ensure mode_id is set in app_model.
        """

        on_str = f" with arg {initial_arg}" if initial_arg is not None else ""   
        logger.debug(f"Entering {self.__class__.__name__} {on_str}")


    def on_leave(self):
        """ Actions to perform when exiting the mode. Override in subclasses if needed. """
        logger.debug(f"Exiting {self.__class__.__name__}")


    def prepare_check_enter(self, on_arg=None) -> object:
        """ Check if the mode can be entered. Prepare and Return initial object. """
        # to be overridden in subclasses if needed.
        return on_arg
    
    @property
    def current_panel(self) -> Data_Panel:
        """ Get the currently visible data panel. """
        return self.stacked_panel.currentWidget() 


    def set_current_panel (self, minimized: bool = False, refresh=True):
        """ Set self and all modes to minimized or normal state. """

        if minimized:
            new_current = self.panel_small
        else:
            new_current = self.panel

        self.stacked_panel.setCurrentWidget (new_current)

        if refresh:
            new_current.refresh(always=True)


    def toggle_minimized (self):
        """ Toggle the minimized state of the data panel. """
        self.sig_toggle_minimized.emit ()



    # ---- User actions ----

    def cancel (self):
        """ User action: Cancel current mode and request exit. """
 
        QTimer.singleShot (0, self.sig_leave_requested.emit)    # leave after current events processed


    def finish (self):
        """ User action: Finish current mode and request exit. """

        QTimer.singleShot (0, self.sig_leave_requested.emit)    # leave after current events processed


    def switch_mode (self, mode_id: Mode_Id, on_arg=None):
        """ User action: Switch to another mode. """
         
        QTimer.singleShot (0, lambda: self.sig_switch_mode_requested.emit (mode_id, on_arg))


class Mode_View (Mode_Abstract):
    """
    Application mode for viewing airfoils without modification capabilities.
    """

    mode_id = Mode_Id.VIEW                       # the id of the mode


    @property
    def panel (self) -> Data_Panel:
        """ lower UI main panel """

        if self._panel is None: 

            l = QHBoxLayout()

            p = Panel_File_View (self, self._app_model, width=250, lazy=True)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            p.sig_modify.connect                (self.modify)
            p.sig_optimize.connect              (self.optimize)
            p.sig_new_as_bezier.connect         (self.new_as_Bezier)
            p.sig_save_as.connect               (self.save_as)
            p.sig_rename.connect                (self.rename)
            p.sig_delete.connect                (self.delete)
            p.sig_delete_temp_files.connect     (self.delete_temp_files)
            p.sig_exit .connect                 (self.exit)
            l.addWidget (p)

            l.addWidget (Panel_Geometry        (self, self._app_model, lazy=True))
            l.addWidget (Panel_Panels          (self, self._app_model, lazy=True))
            l.addWidget (Panel_LE_TE           (self, self._app_model, lazy=True))
            l.addWidget (Panel_Bezier          (self, self._app_model, lazy=True))
            l.addWidget (Panel_Flap            (self, self._app_model, lazy=True))

            self._panel = Data_Panel (self, self._app_model, layout=l)

        return self._panel


    @property
    def panel_small (self) -> Data_Panel:
        """ lower UI view panel - small version with only small panels"""

        if self._panel_small is None:

            l = QHBoxLayout()

            p = Panel_File_View_Small (self, self._app_model, has_head=False, width=250, lazy=True)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            p.sig_modify.connect                (self.modify)
            p.sig_optimize.connect              (self.optimize)
            p.sig_new_as_bezier.connect         (self.new_as_Bezier)
            p.sig_save_as.connect               (self.save_as)
            p.sig_rename.connect                (self.rename)
            p.sig_delete.connect                (self.delete)
            p.sig_delete_temp_files.connect     (self.delete_temp_files)
            p.sig_exit .connect                 (self.exit)
            l.addWidget (p)

            l.addWidget (Panel_Geometry_Small   (self, self._app_model, has_head=False, lazy=True))
            l.addWidget (Panel_Panels_Small     (self, self._app_model, has_head=False, lazy=True))
            l.addWidget (Panel_LE_TE_Small      (self, self._app_model, has_head=False, lazy=True))
            l.addWidget (Panel_Bezier_Small     (self, self._app_model, has_head=False, lazy=True))
            l.addWidget (Panel_Flap_Small       (self, self._app_model, has_head=False, lazy=True))

            self._panel_small = Data_Panel (self, self._app_model, layout=l)

        return self._panel_small


    def prepare_check_enter(self, initial_airfoil: Airfoil | None) -> Airfoil:
        """ Check if the mode can be entered. Prepare and Return initial object. """

        if isinstance (initial_airfoil, Airfoil):
            return initial_airfoil
        else:
            return self._app_model.airfoil


    def on_enter(self, airfoil: Airfoil):

        self._app_model.set_airfoil (airfoil, silent=True, load_settings=True)

        # switch app_model to this mode - remove any case
        self._app_model.set_mode_and_case (self.mode_id, None)

        super().on_enter()

    # ---- User actions ----

    def modify (self): 
        """ switch to modify mode """
        self.switch_mode (Mode_Id.MODIFY)


    def optimize (self): 
        """ switch to optimize mode """
        self.switch_mode (Mode_Id.OPTIMIZE)


    def save_as (self): 
        """ save current airfoil as ..."""

        airfoil = self._app_model.airfoil

        dlg = Airfoil_Save_Dialog (self.stacked_panel, airfoil)
        ok_save = dlg.exec()

        if ok_save: 
            self._app_model.set_airfoil (airfoil)                         # refresh with new

            self._toast_message (f"New airfoil {airfoil.fileName} saved", toast_style=style.GOOD)
            logger.info (f"Airfoil saved as {airfoil.fileName}")


    def rename (self): 
        """ rename current airfoil as ..."""

        airfoil = self._app_model.airfoil

        old_pathFileName = airfoil.pathFileName_abs

        dlg = Airfoil_Save_Dialog (self.stacked_panel, airfoil , rename_mode=True, remove_designs=True)
        ok_save = dlg.exec()

        if ok_save: 

            # delete old one 
            if os.path.isfile (old_pathFileName):  
                os.remove (old_pathFileName)

            # a copy with new name was created 
            self._app_model.set_airfoil (airfoil)

            self._toast_message (f"Airfoil renamed to {airfoil.fileName}", toast_style=style.GOOD)
            logger.info (f"Airfoil renamed to {airfoil.fileName}")


    def delete (self): 
        """ delete current airfoil ..."""

        airfoil = self._app_model.airfoil

        if not os.path.isfile (airfoil.pathFileName_abs): return 

        text = f"Airfoil <b>{airfoil.fileName}</b> including temporary files will be deleted."
        button = MessageBox.warning (self.stacked_panel, "Delete airfoil", text)

        if button == QMessageBox.StandardButton.Ok:

            self.delete_temp_files (silent=True)
            os.remove (airfoil.pathFileName_abs)                               # remove airfoil

            self._toast_message (f"Airfoil {airfoil.fileName} deleted", toast_style=style.GOOD)
            logger.info (f"Airfoil {airfoil.fileName} deleted")

            next_airfoil = get_next_airfoil_in_dir (airfoil, example_if_none=True)
            self._app_model.set_airfoil (next_airfoil)                           # try to set on next airfoil

            if next_airfoil.isExample:
               button = MessageBox.info (self.stacked_panel, "Delete airfoil", "This was the last airfoil in the directory.<br>" + \
                                               "Showing Example airfoil") 


    def delete_temp_files (self, silent=False): 
        """ delete all temp files and directories of current airfoil ..."""

        airfoil = self._app_model.airfoil
        if not os.path.isfile (airfoil.pathFileName_abs): return 

        delete = True 

        if not silent: 
            text = f"All temporary files and directories of Airfoil <b>{airfoil.fileName}</b> will be deleted."
            button = MessageBox.warning (self.stacked_panel, "Delete airfoil", text)
            if button != QMessageBox.StandardButton.Ok:
                delete = False
        
        if delete: 
            Case_Direct_Design.remove_design_dir (airfoil.pathFileName_abs)    # remove temp design files and dir 
            Worker.remove_polarDir (airfoil.pathFileName_abs)                  # remove polar dir 
            Xoptfoil2.remove_resultDir (airfoil.pathFileName_abs)              # remove Xoptfoil result dir 

            if not silent:
                self._toast_message (f"Temporary files of Airfoil {airfoil.fileName} deleted", toast_style=style.GOOD)


    def exit (self):
        """ User action: leave current mode and app. """

        self.sig_exit_requested.emit ()


    def new_as_Bezier (self):
        """ create new Bezier airfoil based on current airfoil, create Case, switch to modify mode """

        self.switch_mode (Mode_Id.AS_BEZIER)




class Mode_Modify (Mode_Abstract):
    """
    Application mode for modifying airfoils.
    """

    mode_id = Mode_Id.MODIFY                       # the id of the mode


    def prepare_check_enter(self, on_arg = None ) -> Airfoil | None:
        """ Check if the mode can be entered. Override in subclasses if needed. """

        # info if airfoil is flapped 
        if self._app_model.airfoil.geo.isProbablyFlapped:

            text = "The airfoil is probably flapped and will be normalized.\n\n" + \
                   "Modifying the geometry can lead to strange results."
            button = MessageBox.confirm (self.stacked_panel, "Modify Airfoil", text)
            if button == QMessageBox.StandardButton.Cancel:
                return None
        return self._app_model.airfoil


    def on_enter(self, airfoil: Airfoil):
        """ Actions to perform when entering the mode. """
        # ensure example airfoil is saved to file to ease consistent further handling in widgets
        if airfoil.isExample:
            airfoil.save()

        # show airfoil design initially
        self._app_model.set_show_airfoil_design (True)

        # switch app_model to this mode with new Design Case - will get/create first design
        self._app_model.set_mode_and_case (self.mode_id, Case_Direct_Design (airfoil))

        super().on_enter()


    def on_leave(self):
        """ Actions to perform when exiting the mode.  """

        super().on_leave()

        if self._app_model.case.airfoil_final:
            next_airfoil = self._app_model.case.airfoil_final       # final airfoil created in finish
        else:
            next_airfoil = self._app_model.case.airfoil_seed
        next_airfoil.set_usedAs (usedAs.NORMAL)                     # normal AE color 

        self._app_model.case.close()                                # shut down case
        self._app_model.set_case (None)
        self._app_model.set_airfoil (next_airfoil, silent=True)     # we'll continue with set airfoil to final or seed


    def cancel(self):
        """ User action: Cancel current mode and request exit. """

        self._app_model.case.set_airfoil_final (None)                 # just sanity

        # rest will be done in on_leave
        super().cancel()


    def finish(self):
        """ User action: Finish current mode and request exit. """

        # create new, final airfoil based on actual design and path from airfoil org 

        case : Case_Direct_Design = self._app_model.case
        new_airfoil = case.get_final_from_design (self._app_model.airfoil)

        # dialog to edit name, choose path, ..

        dlg = Airfoil_Save_Dialog (parent=self.stacked_panel, getter=new_airfoil,
                                   parentPos=(0.25,-1), dialogPos=(0,1))
        ok = dlg.exec()
        if not ok: return                                       # save was cancelled - return to modify mode 

        # set final airfoil in case - rest will be done in on_leave
        self._app_model.case.set_remove_designs_on_close (dlg.remove_designs)
        self._app_model.case.set_airfoil_final (new_airfoil)

        self._toast_message (f"New airfoil {new_airfoil.fileName} saved", toast_style=style.GOOD)
        logger.info (f"New airfoil {new_airfoil.fileName} created from {self._app_model.airfoil.fileName}")

        super().finish()
    

    @property
    def panel (self) -> Data_Panel:
        """ lower UI main panel - modify mode """

        if self._panel is None: 

            l = QHBoxLayout()

            p = Panel_File_Modify (self, self._app_model, width=250, lazy=True)
            p.sig_cancel.connect                (self.cancel)
            p.sig_finish.connect                (self.finish)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            l.addWidget (p)

            l.addWidget (Panel_Geometry        (self, self._app_model, lazy=True))
            l.addWidget (Panel_Panels          (self, self._app_model, lazy=True))
            l.addWidget (Panel_LE_TE           (self, self._app_model, lazy=True))
            l.addWidget (Panel_Flap            (self, self._app_model, lazy=True))
            l.addWidget (Panel_Bezier          (self, self._app_model, lazy=True))
            l.addWidget (Panel_Bezier_Match    (self, self._app_model, lazy=True))

            self._panel    = Data_Panel (self, self._app_model, layout = l)

        return self._panel


    @property
    def panel_small (self) -> Data_Panel:
        """ lower UI view panel - small version"""

        if self._panel_small is None: 

            l = QHBoxLayout()

            p = Panel_File_Modify_Small (self, self._app_model, width=250, lazy=True, has_head=False)
            p.sig_cancel.connect                (self.cancel)
            p.sig_finish.connect                (self.finish)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            l.addWidget (p)

            l.addWidget (Panel_Geometry_Small       (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Panels_Small         (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_LE_TE_Small          (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Bezier_Small         (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Bezier_Match_Small   (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Flap_Small           (self, self._app_model, lazy=True, has_head=False))

            self._panel_small = Data_Panel (self, self._app_model, layout=l)

        return self._panel_small



class Mode_As_Bezier (Mode_Abstract):
    """
    Application mode for create new airfoil based on Bezier curves.
    """

    mode_id = Mode_Id.AS_BEZIER                       # the id of the mode

            
    def prepare_check_enter(self, on_arg = None) -> Airfoil | None:
        """ Check if the mode can be entered. Override in subclasses if needed. """

        if not self._app_model.airfoil.isNormalized:

            text = f"{self._app_model.airfoil.fileName} is not normalized.\n\n" + \
                   "Match Bezier will work on a normalized copy of the airfoil."
            button = MessageBox.confirm (self.stacked_panel, "New As Bezier", text)
            if button == QMessageBox.StandardButton.Cancel:
                return None
        return self._app_model.airfoil


    def on_enter(self, airfoil: Airfoil):

        # ensure example airfoil is saved to file to ease consistent further handling in widgets
        if airfoil.isExample:
            airfoil.save()

        # switch app_model to this mode with new Design Case - will get/create first design
        self._app_model.set_mode_and_case (self.mode_id, Case_As_Bezier (airfoil))

        # show airfoil design initially
        self._app_model.set_show_airfoil_design (True)

        super().on_enter()


    def on_leave(self):
        """ Actions to perform when exiting the mode. Override in subclasses if needed. """

        if self._app_model.case.airfoil_final:
            next_airfoil = self._app_model.case.airfoil_final       # final airfoil created in finish
        else:
            next_airfoil = self._app_model.case.airfoil_seed

        self._app_model.case.close()                                 # shut down case
        self._app_model.set_case (None)
        self._app_model.set_airfoil (next_airfoil, silent=True)      # we'll continue with set airfoil to final or seed

        super().on_leave()


    def cancel(self):
        """ User action: Cancel current mode and request exit. """

        self._app_model.case.set_airfoil_final (None)                 # just sanity

        # rest will be done in on_leave
        super().cancel()


    def finish(self):
        """ User action: Finish current mode and request exit. """

        # create new, final airfoil based on actual design and path from airfoil org 

        case : Case_As_Bezier = self._app_model.case
        new_airfoil = case.get_final_from_design (self._app_model.airfoil)

        # dialog to edit name, choose path, ..

        dlg = Airfoil_Save_Dialog (parent=self.stacked_panel, getter=new_airfoil,
                                   parentPos=(0.25,-1), dialogPos=(0,1))
        ok = dlg.exec()
        if not ok: return                                       # save was cancelled - return to modify mode 

        # set final airfoil in case - rest will be done in on_leave
        self._app_model.case.set_remove_designs_on_close (dlg.remove_designs)
        self._app_model.case.set_airfoil_final (new_airfoil)

        self._toast_message (f"New airfoil {new_airfoil.fileName} saved", toast_style=style.GOOD)
        logger.info (f"New airfoil {new_airfoil.fileName} created from {self._app_model.airfoil.fileName}")

        super().finish()
    

    @property
    def panel (self) -> Data_Panel:
        """ lower UI main panel - modify mode """

        if self._panel is None: 

            l = QHBoxLayout()

            p = Panel_File_Modify (self, self._app_model, width=250, lazy=True)
            p.sig_cancel.connect                (self.cancel)
            p.sig_finish.connect                (self.finish)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            l.addWidget (p)

            l.addWidget (Panel_Geometry        (self, self._app_model, lazy=True))
            l.addWidget (Panel_Panels          (self, self._app_model, lazy=True))
            l.addWidget (Panel_Bezier          (self, self._app_model, lazy=True))
            l.addWidget (Panel_Bezier_Match    (self, self._app_model, lazy=True))

            self._panel    = Data_Panel (self, self._app_model, layout = l)

        return self._panel


    @property
    def panel_small (self) -> Data_Panel:
        """ lower UI view panel - small version"""

        if self._panel_small is None: 

            l = QHBoxLayout()

            p = Panel_File_Modify_Small (self, self._app_model, width=250, lazy=True, has_head=False)
            # p.sig_cancel.connect                (self.cancel)
            p.sig_finish.connect                (self.finish)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            l.addWidget (p)

            l.addWidget (Panel_Geometry_Small       (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Panels_Small         (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Bezier_Small         (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Bezier_Match_Small   (self, self._app_model, lazy=True, has_head=False))

            self._panel_small = Data_Panel (self, self._app_model, layout=l)

        return self._panel_small


# ------------------------------------------------------------------------------


class Mode_Optimize (Mode_Abstract):
    """
    Application mode for optimizing airfoils.
    """

    mode_id = Mode_Id.OPTIMIZE                       # the id of the mode


    def _save_input_file (self, ask = False, toast=True): 
        """ save xo2 input options - optionally ask user"""

        case: Case_Optimize = self._app_model.case

        # check if changes were made in current case 

        if case and case.input_file.isChanged:
            if ask: 
                text = f"Save changes made for <b>{case.name}</b>?"
                button = MessageBox.save (self.stacked_panel, "Save Case", text,
                                          buttons = QMessageBox.StandardButton.Save | 
                                                    QMessageBox.StandardButton.Discard)
                if button == QMessageBox.StandardButton.Discard:
                    return 
                
            saved = case.input_file.save_nml()

            if toast and saved:
                self._toast_message (f"Options saved to Input file", toast_style=style.GOOD)


    def on_enter(self, pathFileName: str):
        """ Actions to perform when entering the mode. """

        # switch app_model to this mode with new Optimize Case - will get/create first design
        self._app_model.set_mode_and_case (self.mode_id, Case_Optimize (pathFileName))

        # don't show airfoil design initially
        self._app_model.set_show_airfoil_design(False)

        super().on_enter()


    def on_leave(self):
        """ Actions to perform when exiting the mode """

        super().on_leave()

        # set next airfoil to final or seed
        if self._app_model.case:
            if self._app_model.case.airfoil_final:
                next_airfoil = self._app_model.case.airfoil_final       # final airfoil created in finish
            else:
                next_airfoil = self._app_model.case.airfoil_seed
            next_airfoil.set_usedAs (usedAs.NORMAL)                     # normal AE color 

            self._app_model.case.close()                                # shut down case
            self._app_model.set_airfoil (next_airfoil, silent=True)     # we'll continue with set airfoil to final or seed
    
        self._app_model.set_case (None)
    

    def prepare_check_enter(self, on_arg=None) -> str:
        """ Check / prepare if the mode can be entered. Get/Create input file for xo2 """

        if not Xoptfoil2.ready : return False

        # an input file is already set - check existence

        if on_arg and Input_File.is_xo2_input(on_arg): 
                return on_arg
        elif on_arg:
            logger.error (f"Cannot enter Optimize mode - provided input file is not xo2 input: {on_arg}")
            return None
            
        # no xo2 file as argument - ask user to select existing or create new one

        diag = Xo2_Select_Dialog (self.stacked_panel, self._app_model, parentPos=(0.4,-0.5), dialogPos=(0,1))
        rc = diag.exec()

        if rc == QDialog.DialogCode.Accepted:

            if diag.input_fileName:                                     
                # existing xo2 file selected (or new version)
                return os.path.join(diag.workingDir, diag.input_fileName)
            else: 
                # create new xo2 file based on current airfoil
                seed_airfoil = self._app_model.airfoil
                workingDir   = seed_airfoil.pathName_abs

                diag = Xo2_New_Dialog (self.stacked_panel, workingDir, seed_airfoil, 
                                       watchdog=self._app_model._watchdog, 
                                       parentPos=(0.5,0.0), dialogPos=(0,1.1))
                rc = diag.exec()

                if rc == QDialog.DialogCode.Accepted:
                    return os.path.join(diag.workingDir, diag.input_fileName)
        return None                             


    @property
    def panel (self) -> Data_Panel:
        """ lower UI main panel"""

        if self._panel is None: 

            l = QHBoxLayout()  

            p = Panel_Xo2_File (self, self._app_model, width=250, lazy=True)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            p.sig_open_next.connect             (self.open_next)
            p.sig_new.connect                   (self.new)
            p.sig_new_version.connect           (self.new_version)
            p.sig_finish.connect                (self.finish)
            l.addWidget (p)
            # p.sig_cancel.connect                (self.cancel)

            l.addWidget (Panel_Xo2_Case                    (self, self._app_model, lazy=True))
            l.addWidget (Panel_Xo2_Operating_Conditions    (self, self._app_model, lazy=True))
            l.addWidget (Panel_Xo2_OpPoint_Defs            (self, self._app_model, lazy=True))
            l.addWidget (Panel_Xo2_Geometry_Targets        (self, self._app_model, lazy=True))
            l.addWidget (Panel_Xo2_Curvature               (self, self._app_model, lazy=True))
            p = Panel_Xo2_Advanced              (self, self._app_model, lazy=True)
            p.sig_edit_input_file.connect       (self.edit_input_file)
            l.addWidget (p)

            self._panel = Data_Panel (self, self._app_model, layout = l)

            self._app_model.sig_xo2_input_changed.connect (self._panel.refresh)

        return self._panel


    @property
    def panel_small (self) -> Data_Panel:
        """ lower UI main panel """

        if self._panel_small is None: 

            l = QHBoxLayout()        

            p = Panel_Xo2_File_Small (self, self._app_model, width=250, lazy=True, has_head=False)
            p.sig_toggle_panel_size.connect     (self.toggle_minimized)
            p.sig_finish.connect                (self.finish)
            l.addWidget (p)

            l.addWidget (Panel_Xo2_Case_Small               (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Xo2_Operating_Small          (self, self._app_model, lazy=True, has_head=False))
            l.addWidget (Panel_Xo2_Geometry_Targets_Small   (self, self._app_model, lazy=True, has_head=False))

            self._panel_small = Data_Panel (self, self._app_model, layout = l)

            self._app_model.sig_xo2_input_changed.connect (self._panel_small.refresh)

        return self._panel_small


    # ---- User actions ----


    def run_xo2 (self): 
        """ run optimizer"""

        self._app_model.run_xo2 ()


    def edit_input_file (self):
        """ slot user action - edit xo2 input file """

        self._save_input_file (ask=False, toast=True)

        case : Case_Optimize = self._app_model.case
        diag = Xo2_Input_File_Dialog (self.stacked_panel, case.input_file, parentPos=(0.8,0,9), dialogPos=(1,1))
        diag.exec () 

        if diag.result() == QDialog.DialogCode.Accepted:
            msg = "Input file successfully checked and saved"
            self._toast_message (msg, toast_style=style.GOOD)
            self._app_model.notify_xo2_input_changed()


    def open_next (self, pathFileName: str):
        """ slot user action - open next xo2 input file """

        self._save_input_file (ask=True)

        # set next file to open
        self.switch_mode (Mode_Id.OPTIMIZE, pathFileName)        


    def new_version (self):
        """ slot user action - create new version of xo2 input file """

        self._save_input_file (ask=True)

        case : Case_Optimize = self._app_model.case

        workingDir   = case.input_file.workingDir
        cur_fileName = case.input_file.fileName
        new_fileName = Input_File.new_fileName_version (cur_fileName, workingDir)

        if new_fileName:

            copyfile (os.path.join (workingDir,cur_fileName), os.path.join (workingDir,new_fileName))
            self.switch_mode (Mode_Id.OPTIMIZE, os.path.join (workingDir,new_fileName))

            self._toast_message (f"New version {new_fileName} created", toast_style=style.GOOD) 
        else: 
            MessageBox.error   (self.stacked_panel,'Create new version', f"New Version of {cur_fileName} could not be created.",
                                min_width=350)


    def new (self):
        """ slot user action - create new xo2 input file """

        self._save_input_file (ask=True)

        seed_airfoil = self._app_model.airfoil_seed
        workingDir   = self._app_model.workingDir

        diag = Xo2_New_Dialog (self.stacked_panel, workingDir, seed_airfoil, 
                               watchdog=self._app_model._watchdog,
                               parentPos=(0.5,0.0), dialogPos=(0.5,1.1))
        rc = diag.exec()

        if rc == QDialog.DialogCode.Accepted:
            pathFileName = os.path.join(diag.workingDir, diag.input_fileName)
            self.switch_mode (Mode_Id.OPTIMIZE, pathFileName)


    def finish (self):
        """ slot user action - finish optimize airfoil - switch back to view mode """ 

        # be sure input file data is written to file 
        self._save_input_file (ask=True)

        # rest will be done in on_leave
        super().finish()




# ------------------------------------------------------------------------------


class Modes_Manager (QObject):
    """
    Manages the different application modes.

    The Modes Manager provides a stacked data panel which presents the lower part of the UI
    according to the current mode.
    """

    sig_close_requested = pyqtSignal()                      # signal to request app close


    def __init__(self, app_model: App_Model):
        super().__init__()

        self._modes_dict  = {}
        self._app_model   = app_model

        self._modes_panel : QStackedWidget = None

        self._height            = 250                       # default height of modes panel
        self._height_minimized  = 150                       # default height of modes panel when minimized

        s = Settings()
        self._is_minimized = s.get('lower_panel_minimized', False)


    def _switch_mode_panel (self, mode: Mode_Abstract):
        """ switch stacked widget to panel of mode """

        if not (mode.mode_id in self._modes_dict):
            logger.error (f"Mode {mode} not registered in Mode_Manager.")

        # set small oder normal  
        mode.set_current_panel (self._is_minimized, refresh=True)

        # switch stacked widget to new mode panel
        self._modes_panel.setCurrentWidget (mode.stacked_panel)

        # setting of the actual width of the modes panel after switching needs to be done
        # after all current events are processed so that the size hint of the new panel is valid
        QTimer.singleShot (100, self._set_min_width)        # set min width after current events processed


    def _set_min_width (self):
        """ set minimum width of modes panel according to current mode panel """

        # because QStackedWidget takes the width of the widest widget of all its children,
        # we need to set the minimum width of the stacked widget to the minimum width of the current mode panel

        if self.current_mode is not None:
            self._modes_panel.setMinimumWidth (0)
            self._modes_panel.adjustSize()
            min_width = self.current_mode.current_panel.calc_min_width()
            self._modes_panel.setMinimumWidth (min_width)


    @property
    def current_mode (self) -> Mode_Abstract:
        """ Get the current active mode. """

        mode_id = self._app_model.mode_id
        if mode_id is None:
            return None
        elif mode_id in self._modes_dict:
            return self._modes_dict[mode_id]
        else:
            logger.warning(f"Mode {mode_id} not registered in Mode_Manager.")
            return None     
       

    def add_mode(self, mode: Mode_Abstract):
        """ add a mode to the manager """

        mode_id = mode.mode_id

        if not mode_id in self._modes_dict:

            self._modes_dict [mode_id] = mode                        # register mode

            # connect to signals of Mode
            mode.sig_exit_requested.connect         (self.exit)
            mode.sig_leave_requested.connect        (self.leave_mode)
            mode.sig_switch_mode_requested.connect  (self.switch_mode)
            mode.sig_toggle_minimized.connect       (self.toggle_minimized)


    def set_mode (self, mode_id: Mode_Id, on_arg=None):
        """ set initial mode """

        if not mode_id in self._modes_dict:
            logger.error(f"Mode {mode_id} not registered in Mode_Manager.")
            return

        if self._app_model.mode_id is not None:
            logger.error("Initial mode already set in App_Model.")
            return

        mode : Mode_Abstract = self._modes_dict [mode_id]

        # prepare new mode and argument to work on
        arg =  mode.prepare_check_enter (on_arg)

        mode.on_enter(arg)                                              # set mode
        mode.set_current_panel (self._is_minimized)                     # select the right panels small or not



    def stacked_modes_panel (self) -> QStackedWidget:
        """ build all modes panels and return them as stacked widget """

        if self._modes_panel is not None:
            return self._modes_panel
        
        self._modes_panel = QStackedWidget()

        # collect all modes and their panels
        mode : Mode_Abstract
        for mode_id, mode in self._modes_dict.items():
            self._modes_panel.addWidget (mode.stacked_panel)                 # add mode's data panels to stacked widget

        # switch stacked widget to current mode panel
        if self.current_mode is not None:
            self._modes_panel.setCurrentWidget (self.current_mode.stacked_panel)

        return self._modes_panel


    def switch_mode (self, new_mode_id: Mode_Id, on_arg=None):
        """ switch to given mode """

        if not new_mode_id in self._modes_dict:
            logger.error(f"Mode {new_mode_id} not registered in Mode_Manager.")
            return

        if self._app_model.mode_id is None:
            logger.error("Cannot switch modes: No initial mode set")
            return

        # leave current mode
        self.current_mode.on_leave()

        # prepare new mode and argument to work on
        new_mode   : Mode_Abstract = self._modes_dict [new_mode_id]
        new_arg =  new_mode.prepare_check_enter (on_arg)
        if new_arg is None:  return                                         # cannot enter mode         
        
        # enter new mode, switch data panels
        new_mode.on_enter(new_arg)                                          # prepare enter new mode
        self._switch_mode_panel (new_mode)                                  # switch stacked widget - make visible



    def leave_mode (self):
        """ leave current mode and return to view mode """

        self.switch_mode (Mode_Id.VIEW)


    def exit(self):
        """ exit mode and close app if in view mode """

        if self.current_mode is not None:
            self.current_mode.on_leave()

        s = Settings()
        s.set('lower_panel_minimized', self._is_minimized)
        s.save()

        self.sig_close_requested.emit()             # close app if view mode finished


    def set_height (self, height: int, minimized: int|None = None):
        """ set height of modes panel """

        self._height = height
        self._height_minimized = minimized if minimized is not None else self._height_minimized

        self._set_minimized (self._is_minimized)                     # apply height change


    def _set_minimized (self, minimized: bool):
        """ set minimized state of modes panel """

        self._is_minimized = minimized

        if self.current_mode is not None:
            self.current_mode.set_current_panel (minimized)         # set panel small or normal

        # set to predefined height
        height = self._height_minimized if minimized else self._height
        self._modes_panel.setFixedHeight (height)    


    def toggle_minimized (self):
        """ toggle minimized state of modes panel """

        if self.current_mode is not None:
            self._set_minimized (not self._is_minimized)

            # adjust new min with of data panel
            self._set_min_width()        





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""  

Diagram (items) for xo2

"""

from typing                 import override

import pyqtgraph as pg                                

from ..base.artist          import Artist
from ..base.diagram         import Diagram, Diagram_Item 
from ..model.polar_set      import var
from ..model.xo2_results    import Optimization_History_Entry

from .xo2_artists           import Xo2_Design_Radius_Artist, Xo2_Improvement_Artist, Xo2_OpPoint_Defs_Artist
from .ae_diagrams           import Item_Airfoil, Item_Polars
from .ae_artists            import Polar_Artist

from ..app_model            import App_Model


import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class Diagram_Xo2_Airfoil_and_Polar (Diagram):
    """ Little Diagram with Airfoil and Polar Item used for new xo2 optimization case"""

    width  = (None, None)               # (min,max) 
    height = (400, None)                # (min,max)


    def __init__(self, *args,  **kwargs):
        super().__init__(*args, **kwargs)

        self.graph_layout.setContentsMargins (5,10,5,5)  # default margins

        # show opPoint definitions
        for artist in self._get_artist (Xo2_OpPoint_Defs_Artist):
            artist.set_show (True)
            artist.set_show_mouse_helper (False) 

        # switch off polar legend 
        for artist in self._get_artist (Polar_Artist):
            artist.set_show_legend (False) 

    @property 
    def app_model (self) -> App_Model:
        """ application model"""
        return self.dataObject()


    @override
    def create_diagram_items (self):
        """ create all plot Items and add them to the layout """

        r = 0
        item = Item_Airfoil (self,self.app_model)
        item.setMinimumSize (300, 200)
        self._add_item (item, r, 0, colspan=2)

        r += 1
        default_settings = [{"xyVars" : (var.CD,var.CL)}, {"xyVars" : (var.CL,var.GLIDE)}]

        for iItem in [0,1]:
            # create Polar items with init values vor axes variables 
            item = Item_Polars (self, self.app_model, show=True)
            item.name = f"{Item_Polars.name}_{iItem+1}"                 # set unique name as there a multiple items
            item._set_settings (default_settings[iItem])                        # set default settings first
            self._add_item (item, r, iItem, rowStretch=3)

        self.graph_layout.setRowStretchFactor (0,2)
        self.graph_layout.setRowStretchFactor (1,3)


    @override
    def create_view_panel (self):
        """ no view_panel"""
        pass



class Diagram_Item_Design_Radius (Diagram_Item):
    """ 
    Diagram (Plot) Item to plot design radius during optimization
    """

    title       = "Design Radius"
    subtitle    = None                

    min_width   = 100                                    # min size needed - see below 
    min_height  = 100 

    show_buttons = False                                 # no buttons
    show_coords  = False                                 # no coordinates

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setContentsMargins ( 0,10,5,0)


    @property
    def steps (self) -> list ['Optimization_History_Entry']: 
        return self._dataObject ()


    @override
    def plot_title(self):
        super().plot_title (title=self.title, title_size=8, title_color=Artist.COLOR_LEGEND,offset= (10,-10))


    @override
    def setup_artists (self):
        """ create and setup the artists of self"""
        self._add_artist (Xo2_Design_Radius_Artist (self, lambda: self.steps))


    @override
    def setup_viewRange (self):
        """ define view range of this plotItem"""
        axis : pg.AxisItem = self.getAxis ('left')
        axis.setWidth (5)
        axis : pg.AxisItem = self.getAxis ('bottom')
        axis.setHeight (5)
        self.viewBox.enableAutoRange (axis == 'xy')
        self.viewBox.setDefaultPadding (0.0)



class Diagram_Item_Improvement (Diagram_Item):
    """ 
    Diagram (Plot) Item to plot improvement during optimization
    """

    title       = "Improvement"
    subtitle    = None                

    min_width   = 100                                    # min size needed - see below 
    min_height  = 100 

    show_buttons = False                                 # no buttons
    show_coords  = False                                 # no coordinates

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setContentsMargins ( 0,10,5,0)

    @property
    def steps (self) -> list ['Optimization_History_Entry']: 
        return self._dataObject ()


    @override
    def plot_title(self):
        super().plot_title (title=self.title, title_size=8, title_color=Artist.COLOR_LEGEND,offset= (10,-10))


    @override
    def setup_artists (self):
        """ create and setup the artists of self"""
        self._add_artist (Xo2_Improvement_Artist (self, lambda: self.steps))


    @override
    def setup_viewRange (self):
        """ define view range of this plotItem"""
        axis : pg.AxisItem = self.getAxis ('left')
        axis.setWidth (5)
        axis : pg.AxisItem = self.getAxis ('bottom')
        axis.setHeight (5)
        self.viewBox.enableAutoRange (axis == 'xy')
        self.viewBox.setDefaultPadding (0.0)



class Diagram_Xo2_Progress (Diagram):
    """ Diagram with design radius and improvement development  """

    _width  = (None, None)               # (min,max) 
    _height = (140, None)                # (min,max)


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.graph_layout.setContentsMargins (5,10,5,5)  # default margins


    @property
    def steps (self) -> list ['Optimization_History_Entry']:
        """ optimization steps imported (up to now)""" 
        return self._dataObject ()
    

    def create_diagram_items (self):
        """ create all plot Items and add them to the layout """
        item = Diagram_Item_Design_Radius (self, lambda: self.steps)
        self._add_item (item, 0, 0)
        item = Diagram_Item_Improvement   (self, lambda: self.steps)
        self._add_item (item, 0, 1)


    @override
    def create_view_panel (self):
        """ no view_panel"""
        pass




#!/usr/bin/env pythonupper
# -*- coding: utf-8 -*-

"""  

The "Artists" to plot a airfoil object on a pg.PlotItem 

"""

from PyQt6.QtGui                import QColor, QPainterPath, QTransform
from PyQt6.QtCore               import Qt, pyqtSignal, QRectF

from ..base.artist              import *
from ..base.common_utils        import *

from ..model.polar_set          import * 
from ..model.xo2_input          import (OpPoint_Definition, OpPoint_Definitions, OPT_TARGET, OPT_MAX, OPT_MIN)
from ..model.xo2_results        import OpPoint_Result, Optimization_History_Entry

from .ae_artists                import _color_airfoil, SYMBOL_TRANSITION_RIGHT


import logging
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# -------- helper functions ------------------------

def _size_opPoint (weighting : float, normal_size = 13) -> float:
    """ returns plot size of an opPoint (def) depending on weighting"""

    # weighting can be negative (meaning fixed weighting - no dynamic)
    return abs(weighting) ** 0.8 * normal_size 



# -------- Movable Points  ------------------------


class Movable_OpPoint_Def (Movable_Point):
    """ 
    Movable opPoint definition
    """

    name = "OpPoint Definition"

    COLOR = QColor ("darkturquoise")

    # -- special symbols for min/max optTypes 
    #
    #               .--x
    #               |
    #               y
    #

    @staticmethod
    def _symbol_opt_up (width=1.0, height1=0.3, height2=0.5):
        path = QPainterPath()

        path.addRect(QRectF(-width / 2, -height1 / 2, width, height1))
        # verticial line
        path.moveTo(0.0, -height1 / 2)
        path.lineTo(0.0, -height2 * 2)
        #triangle
        path.moveTo(-width / 2, -height2 * 2)
        path.lineTo( width / 2, -height2 * 2)
        path.lineTo(       0.0, -height2 * 2 - height2)
        path.lineTo(-width / 2, -height2 * 2)

        return path

    tr : QTransform = QTransform()
    tr.rotate(-90)                                  # because y is downward

    SYMBOL_OPT_UP    = _symbol_opt_up ()                       
    SYMBOL_OPT_LEFT  = tr.map (SYMBOL_OPT_UP)                         
    SYMBOL_OPT_DOWN  = tr.map (SYMBOL_OPT_LEFT)                        
    SYMBOL_OPT_RIGHT = tr.map (SYMBOL_OPT_DOWN)                        


    def __init__ (self, pi : pg.PlotItem, 
                    opPoint_def : OpPoint_Definition,
                    xyVars : Tuple[var, var], 
                    movable = True, 
                    on_selected = None,
                    on_delete   = None,
                    **kwargs):

        self._pi = pi
        self._opPoint_def = opPoint_def
        self._xyVars = xyVars

        self._callback_selected = on_selected if (callable(on_selected) and movable) else None
        self._callback_delete   = on_delete if (callable(on_delete) and movable) else None
        self._highlight_item    = None                       

        brush = QColor ("black")
        brush.setAlphaF (0.3) 

        size    = _size_opPoint (opPoint_def.weighting)
        xy      = self.xy_in_xyVars()                                   # get x,y coordinates in xyVars

        # sanity - opPoint_def could be not ready because seed polar has to be calculated async
        if xy == (None, None):
            return 

        movable = movable and self.is_movable_in_xyVars()               # can be moved in xyVars?

        super().__init__(xy, movable=movable, label_anchor = (0, 0.5),
                            color=QColor(self.COLOR), brush=brush,
                            symbol=self._symbol(), size=size, show_label_static = True, **kwargs)

        self.sigShiftClick.connect  (self._delete_opPoint_def)


    def is_movable_in_xyVars (self) -> bool:
        """ returns True if self can be moved in xyVars"""

        opPoint_def = self._opPoint_def

        if opPoint_def.specVar in self._xyVars and opPoint_def.optVar in self._xyVars:
            return True 
        elif opPoint_def._xyVars_are_indirect (self._xyVars):
            return True
        if opPoint_def.specVar in self._xyVars and opPoint_def.optType in [OPT_MIN, OPT_MAX]:
            return True 
        else:
            return False
    

    def _symbol (self) -> QPainterPath | str:
        """ returns the symbol for self depending on optType"""

        xVar, yVar = self._xyVars

        optVar, optType = self._opPoint_def.optVar_type_for_xyVars (self._xyVars)

        if optType == OPT_TARGET:
            s = 's'
        elif optType == OPT_MAX:
            if optVar == yVar:
                s = self.SYMBOL_OPT_UP
            else:
                s = self.SYMBOL_OPT_RIGHT
        elif optType == OPT_MIN:
            if optVar == yVar:
                s = self.SYMBOL_OPT_DOWN
            else:
                s = self.SYMBOL_OPT_LEFT
        else:                                           # either optVar nor specVar are in diagram
            s = 's'
        
        return s


    def _delete_opPoint_def (self, _):
        """ shift-click - delete me opPoint_dev"""

        if callable(self._callback_delete):
            # callback / emit signal delayed so we leave the scope of Graphics 
            QTimer() .singleShot(0, lambda: self._callback_delete (self._opPoint_def))   


    def xy_in_xyVars (self) -> tuple: 
        """ returns the x,y coordinates of self in xyVars"""

        polar_point = self._opPoint_def.polar_point ()
        if polar_point is None:
            return None, None
        
        x= polar_point.get_value(self._xyVars[0])
        y= polar_point.get_value(self._xyVars[1])     
        if x is None or y is None:
            return None, None       
        
        return x,y 



    @override
    def mouseClickEvent(self, ev):
        """ handle callback to parent when clicked with left button to select self """

        if not self.moving and ev.button() == Qt.MouseButton.LeftButton  \
                           and not (ev.modifiers() & Qt.KeyboardModifier.ShiftModifier): # shift is handled in super
            if callable(self._callback_selected):
                ev.accept()

                # callback / emit delayed so we leave the scope of Graphics 
                QTimer() .singleShot(0, lambda: self._callback_selected (self._opPoint_def))   
        else:
            super().mouseClickEvent (ev)


    @override
    def _moving (self, _):
        """ slot -point is moved"""
        self._opPoint_def.set_xyValues_for_xyVars (self._xyVars, (self.x,self.y))

        self.setPos (self.xy_in_xyVars())                      # ... if we run against limits 

        # also refresh pos of highlight item 
        if isinstance (self._highlight_item, pg.TargetItem):
            self._highlight_item.setPos (self.pos())


    @override
    def label_static (self, *_) -> str:
        """ the static label - can be overloaded """
        return f"{self._opPoint_def.iPoint}" 

    @override
    def label_moving (self, *_):
        """ label for hovering or moving"""
        return f"{self._opPoint_def.labelLong_for (self._xyVars, fix=False)}"

    @override
    def _finished (self, _):
        """ slot - point moving is finished"""

        if callable(self._callback_selected):
            # ensure self will be selected after moving 
            QTimer() .singleShot (0, lambda: self._callback_selected (self._opPoint_def)) 

        if callable(self._callback_changed):
            # callback / emit  delayed so we leave the scope of Graphics 
            QTimer() .singleShot (0, self._callback_changed )   


    def set_highlight_item (self, aItem: pg.TargetItem):
        """ a point item used to highlight current"""

        self._highlight_item = aItem
        if self._highlight_item:
            self._highlight_item.setPos (self.pos())




# -------- Artists ------------------------


class Xo2_OpPoint_Defs_Artist (Artist):
    """ Plot / Modify Xoptfoil2 operating point definitions """

    sig_opPoint_def_changed     = pyqtSignal ()                       # opPoint_def changed  
    sig_opPoint_def_selected    = pyqtSignal (OpPoint_Definition)     # opPoint_def selected 

    def __init__ (self, *args,
                  cur_opPoint_def_fn = None,   
                  isRunning_fn = None, 
                  xyVars = (var.CD, var.CL), 
                  **kwargs):

        super().__init__(*args, **kwargs)

        self._cur_opPoint_def_fn = cur_opPoint_def_fn           # method to get current opPoint definition
        self._isRunning_fn = isRunning_fn
        self._xyVars = xyVars                                   # definition of x,y axis


    @property
    def xyVars(self): return self._xyVars
    def set_xyVars (self, xyVars: Tuple[var, var]): 
        """ set new x, y variables for polar """
        self._xyVars = xyVars 
        self.refresh()


    @property
    def opPoint_defs (self) -> OpPoint_Definitions: 
        return self.data_list
    
    @property
    def cur_opPoint_def (self) -> OpPoint_Definition | None: 
        return self._cur_opPoint_def_fn() if callable (self._cur_opPoint_def_fn) else None

    @property
    def optimizer_isRunning (self) -> bool:
        """ True if optimizer is ready - definitions can be changed"""
        return self._isRunning_fn() if callable (self._isRunning_fn) else False


    def _plot (self): 

        for opPoint_def in self.opPoint_defs:

            pt = Movable_OpPoint_Def  (self._pi, opPoint_def, self.xyVars, 
                                            movable=not self.optimizer_isRunning and self.show_mouse_helper,
                                            on_changed =self.sig_opPoint_def_changed.emit,
                                            on_delete  =self._on_delete,
                                            on_selected=self.sig_opPoint_def_selected.emit)

            if pt.xy_in_xyVars() != (None, None):                   # sanity - if not None, it is in the view

                self._add (pt, name = pt.name_for_legend) 

                # highlight current opPoint def for edit with a big circle 

                if opPoint_def == self.cur_opPoint_def and not self.optimizer_isRunning:

                    brush = QColor (Movable_OpPoint_Def.COLOR)
                    brush.setAlphaF (0.3)
                    highlight_item = self._plot_point (0.02,0.2, color="black", size=60, brush=brush)

                    pt.set_highlight_item (highlight_item)


        # show mouse helper message
        msg = "OpPoint: click to select, shift-click to remove,  ctrl-click to add"
        self.set_help_message (msg)

        # make scene clickable to opPoint:def - delayed as during init scene is not yet available
        QTimer().singleShot (10, self._connect_scene_mouseClick)


    def _connect_scene_mouseClick (self): 
        """ connect mouse click in scene to slot"""           

        scene : pg.GraphicsScene = self._pi.scene()
        if scene:  scene.sigMouseClicked.connect (self._scene_clicked)


    @override
    def _remove_plots(self):
        """ overloaded to disconnect older slot when doing refresh"""

        super()._remove_plots()

        scene : pg.GraphicsScene = self._pi.scene()

        try:                            # slot could be not connected up to now
            scene.sigMouseClicked.disconnect (self._scene_clicked)
        except:
            pass


    def _scene_clicked (self, ev : MouseClickEvent):
        """ 
        slot - mouse click in scene of self - handle add opPoint_def with ctrl-click 
        """ 

        # handle only ctrl-click
        if not (ev.modifiers() & Qt.KeyboardModifier.ControlModifier): return  
       
        # was the scene click in my viewbox?
        if isinstance(ev.currentItem, pg.ViewBox):
            viewbox : pg.ViewBox = ev.currentItem
        else:
            viewbox : pg.ViewBox = ev.currentItem.getViewBox()

        if viewbox == self._pi.vb:                     # is this my view box (or of another plot item)? 
            
            # get scene coordinates of click pos and map to view box 
            ev.accept()
            pos : pg.Point = viewbox.mapSceneToView(ev.scenePos())

            # create new opPoint_def in the list of opPoint_defs 
            new_opPoint_def = self.opPoint_defs.create_in_xyVars (self.xyVars, pos.x(), pos.y())

            if new_opPoint_def is not None:
                self.sig_opPoint_def_changed.emit()
                self.sig_opPoint_def_selected.emit (new_opPoint_def)

            else: 
                self._plot_text ('New OpPoint Definition cannot be created in this diagram', color=COLOR_ERROR, itemPos=(0.5,0.5))
                QTimer().singleShot (1000, self.refresh)            # remove after 1 second


    def _on_delete (self, opPoint_def : OpPoint_Definition):
        """ callback of Movable Point to delete opPoint_def"""

        self.opPoint_defs.delete (opPoint_def)

        self.sig_opPoint_def_changed.emit ()


class Xo2_OpPoint_Artist (Artist):
    """ Plot Xoptfoil2 operating point results """

    def __init__ (self, *args, 
                  opPoint_defs_fn : Callable = None,
                  opPoint_results_fn : Callable = None,
                  prev_opPoint_results_fn : Callable = None,
                  xyVars = (var.CD, var.CL), 
                  **kwargs):

        self._xyVars = xyVars                                       # definition of x,y axis
        self._opPoint_defs_fn = opPoint_defs_fn                     # method to get opPoint definitions
        self._opPoint_results_fn = opPoint_results_fn               # method to get opPoint results
        self._prev_opPoint_results_fn = prev_opPoint_results_fn     # method to get previous opPoint results 

        super().__init__ (*args, **kwargs)

    @property
    def xyVars(self): return self._xyVars
    def set_xyVars (self, xyVars: Tuple[var, var]): 
        """ set new x, y variables for polar """
        self._xyVars = xyVars 
        self.refresh()


    def _plot (self): 
        """ plot all opPoints"""

        opPoints :      list[OpPoint_Result] = self._opPoint_results_fn() 
        if not opPoints: return 

        opPoint_defs :  list[OpPoint_Definition] = self._opPoint_defs_fn() 
        prev_opPoints : list[OpPoint_Result] = self._prev_opPoint_results_fn() [:] 

        legend_name = f"Op Point Result - Design {opPoints[0].idesign}"

        textFill = QColor ("black")
        textFill.setAlphaF (0.5)
       
        for iop, opPoint in enumerate (opPoints):

            prev_opPoint = prev_opPoints[iop] if prev_opPoints else None
            x = opPoint.get_value (self.xyVars[0])
            y = opPoint.get_value (self.xyVars[1])

            if x is None or y is None:
                logger.debug(f"OpPoint {iop} has no coordinates for this diagram - skipping")
                continue

            # get opPoint definition for this opPoint - if not available, use None
            # sanity - opPoint def could be deleted in the meantime 
            if iop < len(opPoint_defs):
                opPoint_def = opPoint_defs[iop]
                if opPoint_def.specValue != opPoint.get_value(opPoint_def.specVar):
                    # logger.warning(f"OpPoint {iop} has no definition with same specValue - skipping")
                    opPoint_def = None
            else:
                opPoint_def = None


            color  = self._opPoint_color  (opPoint, opPoint_def)
            symbol = self._opPoint_symbol (opPoint, prev_opPoint)
            label  = self._opPoint_label  (opPoint, opPoint_def, self.xyVars)
            size   = _size_opPoint (opPoint.weighting)

            textColor = QColor (color).darker(130)

            self._plot_point ((x,y), color=color, size=size, symbol=symbol, zValue=20, name=legend_name,
                              text=label, anchor=(0, 0.5), textOffset = (8,0), textColor=textColor, textFill=textFill)

            legend_name = None 

        # message if dynamic weighting of opPoints occurred 

        if any (op.is_new_weighting for op in opPoints):
            self._plot_text ('Dynamic weightings applied', color= "dimgray", fontSize=self.SIZE_HEADER, itemPos=(0.5, 1))


        
    def _opPoint_color (self, opPoint : OpPoint_Result, opPoint_def : OpPoint_Definition | None) -> QColor:
        """ color of opPoint depending on % deviation """

        # sanity 
        if opPoint_def is None: 
            return COLOR_OK

        deviation = opPoint.deviation

        optType = opPoint_def.optType
        optVar  = opPoint_def.optVar
        allow_improved = opPoint_def._myList.allow_improved_target

        if optType == OPT_TARGET:                            # targets - take deviation to target
            # Xoptfoil2 - colors for deviation 
            #
            #   if (op_spec%allow_improved_target .and. opt_type == 'target-drag' .and. dev < 0d0) then 
            #     how_good = Q_GOOD
            #   else if (op_spec%allow_improved_target .and. opt_type /= 'target-drag' .and. dev > 0d0) then 
            #     how_good = Q_GOOD
            #   else
            #     how_good = r_quality (abs(dev), 0.1d0, 2d0, 10d0)      ! in percent
            #   end if
            if allow_improved   and optVar == var.CD and deviation < 0.0:
                color = COLOR_GOOD
            elif allow_improved and optVar != var.CD and deviation > 0.0:
                color = COLOR_GOOD
            else:
                deviation = abs(deviation)
                if deviation < 0.1:
                    color = COLOR_GOOD
                elif deviation >= 10:
                    color = COLOR_ERROR
                elif deviation >= 2.0:
                    color = COLOR_WARNING
                elif deviation >= 0.1:
                    color = COLOR_OK
                else:
                    color = COLOR_OK
        else:                                               # min/max - deviation is improvement 
            # Xoptfoil2 - colors for improvement 
            #
            # if (improv <= 0d0) then 
            #     how_good = Q_BAD
            # elseif (improv < 5d0) then 
            #     how_good = Q_OK
            # elseif (improv >= 5d0) then 
            #     how_good = Q_GOOD
            # else
            #     how_good = Q_BAD
            # end if 
            if optType == OPT_MAX:                          # eg. OPT_MAX of 'glide' - more is better 
                improv = deviation
            else:                                           # eg OPT_MIN of 'cd' - less is better
                improv = deviation

            if improv <= -0.1:
                color = COLOR_WARNING
            elif improv < 2.0:
                color = COLOR_OK
            elif improv >= 2.0:
                color = COLOR_GOOD
            else:
                color = COLOR_ERROR
        return color 


    def _opPoint_symbol (self, opPoint : OpPoint_Result, prev_opPoint : OpPoint_Result | None ):
        """ a triangle in the direction of value change """

        if prev_opPoint is None: return 'o'

        x = opPoint.get_value (self.xyVars[0])
        y = opPoint.get_value (self.xyVars[1])
        prev_x = prev_opPoint.get_value (self.xyVars[0])
        prev_y = prev_opPoint.get_value (self.xyVars[1])

        if prev_x > x:
            symbol = 't3'               # triangle to left
        elif prev_x < x:
            symbol = 't2'               # triangle to right
        elif prev_y > y:
            symbol = 't'                # triangle to down
        elif prev_y < y:
            symbol = 't1'               # triangle to up
        else: 
            symbol = 'o'
        return symbol

        
    def _opPoint_label (self, opPoint : OpPoint_Result, opPoint_def : OpPoint_Definition | None, xyVars: tuple) -> str:
        """ label of opPoint depending on % deviation """

        # sanity 
        if opPoint_def is None: return None

        label   = None 
        optVar  = opPoint_def.optVar
        optType = opPoint_def.optType
        allow_improved = opPoint_def._myList.allow_improved_target

        # optVar must be in diagram to show distance as it comes from Xo2 results 

        if optVar in xyVars:  

            # set distance = 0 if 'allow_improved_target' and result (distance) is better

            distance = opPoint.distance

            if optType == OPT_TARGET:                            # targets - take deviation to target
                if allow_improved   and optVar == var.CD and distance < 0.0:
                    distance = 0.0
                elif allow_improved and optVar != var.CD and distance > 0.0:
                    distance = 0.0
                else:
                    distance = abs(distance)

            # format value for label 

            if  optType == OPT_TARGET:                              # targets - take deviation to target
                if optVar == var.CD and round_down (distance,5):
                    label = f"delta {optVar}: {distance:.5f}"
                elif optVar != var.CD and round_up (distance,2):
                    label = f"delta {optVar}: {distance:.2f}"        
                else: 
                    label = f"hit"        
            else:                                               # min/max - deviation is improvement 
                if  opPoint_def.optType == OPT_MAX:             # eg. OPT_MAX of 'glide' - more is better 
                    improv = distance
                else:                                           # eg OPT_MIN of 'cd' - less is better
                    improv = distance

                if improv: 
                    label = f"d{var.CD} {improv:.5f}" if optVar == var.CD else f"d{var.CD} {improv:.2f}"

        # add flap angle if flap optimize 

        if opPoint_def.flap_optimize:
            if label:
                label = f"{label}, F{opPoint.flap:.1f}°"
            else:
                label = f"F{opPoint.flap:.1f}°"

        return label 
        



class Xo2_Design_Radius_Artist (Artist):
    """ Plot Xoptfoil2 design radius during optimization """

    
    @property
    def steps (self) -> list[Optimization_History_Entry]: 
        """ optimization step entries up to now """
        return self.data_object

    def _plot (self): 

        radius_list = [entry.design_radius for entry in self.steps]

        if radius_list:
            p = pg.PlotDataItem  (radius_list)
            self._add (p)

            # plot last as point 
            x = len(radius_list) - 1
            y = radius_list[-1]
            back_color = QColor ("black")
            back_color.setAlphaF (0.5)

            self._plot_point (x,y, size=5, color=COLOR_OK, text = f"{y:.3f}", textFill=back_color, anchor= (1.1, 0.9))



class Xo2_Improvement_Artist (Artist):
    """ Plot Xoptfoil2 improvement during optimization """
    
    @property
    def steps (self) -> list[Optimization_History_Entry]: 
        """ optimization step entries up to now """
        return self.data_object  

    def _plot (self): 

        improvement_list = [entry.improvement for entry in self.steps]

        if improvement_list:

            p = pg.PlotDataItem  (improvement_list)
            self._add (p)

            # plot last as point 
            x = len(improvement_list) - 1
            y = improvement_list[-1]
            back_color = QColor ("black")
            back_color.setAlphaF (0.5)
            
            self._plot_point (x,y, size=5, color=COLOR_GOOD, text = f"{y:.2f}%", textFill=back_color, 
                              textOffset=(-5,0), anchor= (0.8, -0.1))




class Xo2_Transition_Artist (Artist):
    """ Plot Xoptfoil2 point of transition on design airfoil """

    def __init__ (self, *args, 
                  opPoints_result_fn : Callable = None,
                  **kwargs):

        self._opPoints_result_fn = opPoints_result_fn                  # method to get opPoints of designs

        super().__init__ (*args, **kwargs)


    def _plot (self): 
        """ plot all opPoints"""

        airfoil : Airfoil = self.data_object
        opPoints_result : list[OpPoint_Result ]= self._opPoints_result_fn() 

        if not opPoints_result or not airfoil: return 

        legend_name = "Point of Transition xtr"

        color  = _color_airfoil ([], airfoil).darker (120) 
        size   = 10
        brush  = QColor ("black")
        fill   = QColor ("black")
        fill.setAlphaF (0.5) 

        for side in [airfoil.geo.lower, airfoil.geo.upper]:     

            for iop, opPoint in enumerate (opPoints_result):

                x = opPoint.xtrt if side.isUpper else opPoint.xtrb
                y = side.yFn (x) 
                y = y + 0.008 if side.isUpper else y - 0.008

                text   = f"{iop+1}"
                anchor = (0.5, 1.2) if side.isUpper else (0.5, -0.2)

                self._plot_point (x,y, color=color, symbol=SYMBOL_TRANSITION_RIGHT, size=size, brush=brush, 
                                  text=text, textColor=color, anchor=anchor, textFill=fill, name=legend_name)

                legend_name = None                                      # only once for legend 


    @override
    def _add_legend_item (self, plot_item, name : str = None):
        """ add legend item having 'name'"""

        # non standard legend size of transition symbol 
        if self._pi.legend is not None and self.show_legend and name:

            if isinstance (plot_item,  pg.TargetItem):

                # create a dummy PlotItem as TargetItem won't appear in legend 
                size_legend = 30    
                pen    = plot_item.pen
                symbol = plot_item._path 
                brush  = plot_item.brush
                p = pg.ScatterPlotItem ([], [], pen= pen, brush=brush, symbol=symbol, size=size_legend, pxMode=True)
                self._add (p, name=name) 

            else: 

                super()._add_legend_item (plot_item, name=name)
     
